function demo7( image_name )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FUNCTION DEMO( IMAGE_NAME )
%          IMAGE_NAME is the name of the interested image, 
%          e.g. sid-06-02
%          To run this program, the CD-disc of BA image 
%          database must be mounted on the local computer. 
% Aim: This program is written for demonstration.
%      a. Display the colour images and reduced hand label image.
%      b. Calculate the features.
%      c. MLP classification, display the result. 
%      d. Propagating the scaled likelihood, display 
%         the classification results.
%
%  Copyright  1999  (c)     Xiaojuan Feng  
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



  len=length(image_name);
  fname= image_name((len-8):len);
  	      
  %figure  
  set(0, 'Units', 'pixels')
  sw = (1-0.05)/4;  
  
 % Display the image and labelimage
  disp('Obtaining the image from image database .....');
  ch = strcat('cp /', image_name(2:len), '.col.hips .');
  unix(ch);
  axes('Position',[0.05, 3*sw+0.05, sw-0.05, sw-0.05,]);
  images=hips2image(fname);
  image(images/255);  
  axis off
  title(fname);
  drawnow 
 % Display origianl lable image
  disp(' ');
  disp('Obtaining labeled image from image database  .....');  
  axes('Position',[0.05+sw+0.025, 3*sw+0.05, sw-0.05, sw-0.05,]); 
  ch = strcat('cp /', image_name(2:len), '.obj.def .');
  unix(ch);
  ch = strcat('cp /', image_name(2:len), '.obj .');
  unix(ch);
  labels=label2image(fname);
  labels=labels';
  display_label7(labels); 
  axis off
  drawnow 
  title('Original label image');
  % Display reduced  lable image
  axes('Position',[0.05+2*sw+0.025, 3*sw+0.05, sw-0.05, sw-0.05,]); 
  labelR = reduce_label_1(labels);
  display_label7(labelR);
  axis off
  title('Reduced label image');
  drawnow 
  % Display  colour key  
  axes('Position',[0.05+3*sw+0.025, 3*sw+0.05, sw-0.05, sw-0.05,]);   
  display_colour_key(7);
  drawnow 
       
% Calculate features
  disp(' ');
  disp('Calculating the normalized features ... ');  
  norm_features_1( fname );
  !rm -rf *.obj.def 
  !rm -rf *.obj 
  !rm -rf *.col.hips

% MLP
  disp(' ');
  disp('Classifying by the trained MLP classifer ....') 
  fname(4)='_';
  fname(7)='_';
  [labelMlp, ve] = mlpclassify_1a( fname);
  % Dicsplsy MLP segmentation
  axes('Position',[0.05, 2*sw+0.05, sw-0.05, sw-0.05,]);    
  display_label7( labelMlp );
  axis off
  title('MLP segmentation');
  % Dicsplsy MLP segmentation; diff
  axes('Position',[0.05, sw+0.05, sw-0.05, sw-0.05,]);      
  disp_diff(labelMlp,labelR);
  title('labelMlp - labelR');
  axis off 
  drawnow 
  % Dicsplsy MLP segmentation; entropy
  axes('Position',[0.05, 0.05, sw-0.05, sw-0.05,]);      
  s=size(ve);
  normve = ve ./ (sum(ve,2) * ones(1,s(2)));
  entropyMlp=zeros(128,192);
  entropyMlp1 = -sum(normve .* log2(normve),2);
  entropyMlp = vector2matrix(entropyMlp1, 128, 192);
  maxmlp=2.8074 ;
  data = entropyMlp/maxmlp;
  d(:,:,1)=data; d(:,:,2)=data; d(:,:,3)=data;
  image(d); 
  title('Entropy based on Mlp local prediction');
  axis off 
  drawnow    
  
% Compensating MLP  
  prior_train=[0.1758, 0.1787, 0.0662, ...
                  0.1787, 0.1304, 0.1816, 0.088 ];  
  prior_true=[0.1468, 0.3665, 0.0023, 0.4205, ...
                  0.0491, 0.0078, 0.007];
  prior =  	  prior_true ./ prior_train;
  cve = ve .* ( ones(24576,1)*prior);
  s = size(cve);
 [temp, index] = max( cve,[],2 );
 n=0;
 for ki=1:1:128
        for kj=1:1:192
               n = n+1;
               labelCmlp(ki, kj) = index(n);
        end
     end
   % Dicsplsy CMLP segmentation;
  axes('Position',[0.05+sw+0.025, 2*sw+0.05, sw-0.05, sw-0.05,]);        
  display_label7( labelCmlp ); 
  axis off
  title('Compensated MLP segmentation'); 
   % Dicsplsy CMLP segmentation; diff
  axes('Position',[0.05+sw+0.025, sw+0.05, sw-0.05, sw-0.05]);     
  disp_diff(labelCmlp,labelR);
  axis off
  title('labelCMlp - labelR');
   % Dicsplsy MLP segmentation; entropy
  axes('Position',[0.05+sw+0.025, 0.05, sw-0.05, sw-0.05]);      
  s=size(cve);
  normve = cve ./ (sum(cve,2) * ones(1,s(2)));
  entropyCmlp=zeros(128,192);
  entropyCmlp1 = -sum(normve .* log2(normve),2);
  entropyCmlp = vector2matrix(entropyCmlp1, 128, 192);
  maxmlp=2.8074 ;
  data = entropyCmlp/maxmlp;  
  d(:,:,1)=data; d(:,:,2)=data; d(:,:,3)=data;
  image(d); 
  title('Entropy based on compensated Mlp local  prediction');
  axis off  
  drawnow    
  
% Bayesian networks
  ve = ve ./ ( ones(24576,1)*prior_train);
  disp(' ');
  disp('Classifying by Bayesian network (ML).....')
  axes('Position',[0.05+2*sw+0.025, 2*sw+0.05, sw-0.05, sw-0.05]);        
  labelBnet = classify_1('probsML.dat', ve );
  display_label7( labelBnet);
  axis off
  title('Belief segmentation (ML)')  
  axes('Position',[0.05+2*sw+0.025, sw+0.05, sw-0.05, sw-0.05]) ;
  disp_diff(labelBnet,labelR);
  title('labelBnet-labelR');
  axis off
  drawnow
  % display entropy
  axes('Position',[0.05+2*sw+0.025, 0.05, sw-0.05, sw-0.05]) ;
  load  entropy.dat
  maxE=2.8074;  % for seven classes with  uniform distribution
  data=entropy/maxE;
  d(:,:,1)=data; d(:,:,2)=data; d(:,:,3)=data;
  image(d);  
  title(' Entropy: Belief (ML)');
  axis off
   drawnow
   
% Bayesian networks
  disp(' ');
  disp('Classifying by Bayesian network (CML).....')
  axes('Position',[0.05+3*sw+0.025, 2*sw+0.05, sw-0.05, sw-0.05]) ; 
  labelBnet = classify_1('probsCML.dat', ve );
  display_label7( labelBnet);
  axis off
  title('Belief segmentation (CML)')  
 axes('Position',[0.05+3*sw+0.025, sw+0.05, sw-0.05, sw-0.05]) ;   
  disp_diff(labelBnet,labelR);
  title('labelBnet-labelR');
  axis off
  drawnow
 axes('Position',[0.05+3*sw+0.025, 0.05, sw-0.05, sw-0.05]) ;    
  load  entropy.dat
  maxE=2.8074;  % for seven classes with  uniform distribution
  data=entropy/maxE;
  d(:,:,1)=data; d(:,:,2)=data; d(:,:,3)=data;
  image(d);   
  title(' Entropy: Belief (CML) ');
 axis off  
  unix('rm entropy.dat');




