///////////////////////////////////////////////////////////
//
//   Filename:	hips2image3dCD.cc
//   Aim: 	To read image *.col.hips into Matlab
//
//   Tip:       To compile this source code, first start matlab,
//		then send the following compiling command:
//		>>mex hips2image3dCD.cc -f cxxopts.sh
//              To run the program, type,
//              >>images=hips2image3dCD( filename );
//
//   Xiaojuan Feng  September 1997
//   Modification:
//		14/02/98
///////////////////////////////////////////////////////////
#include  <stdlib.h>
#include  <string.h>
#include  <math.h>
#include  <iostream.h>
#include  <fstream.h>
#include  <math.h>
#include  "mex.h"


//#define  NROW  512
//#define  NCOL  768
#define  NROW  533
#define  NCOL  768
#define FILE_READ_BIN          "rb"
#define LINESIZE	       5000

void  hips2image(double*,char*);
void  error1(char*);
void  file_read_bin(void*,int,int,FILE*,char*);
void  read_pixels_rgb(FILE*, unsigned char**,
                      unsigned char**,unsigned char**);
void  get_line( FILE*, char*, int);
void  read_header(FILE*);
void  read_rgb_image(char*, short int***);


/////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////
void mexFunction(int            nlhs, 
                 mxArray*       plhs[], 
                 int            nrhs, 
                 const mxArray* prhs[])
{
  double* image;
  int  nrow, ncol, buflen, status, dimsv[3];
  char fname[30];
  
  // Check for proper number of arguments 
  
  if (nrhs != 1) {
    mexErrMsgTxt("hips2image requires one input argument: *.hips.");
  } else if ( (nlhs < 1)||(nlhs>1)) {
    mexErrMsgTxt("hips2image requires one output argument.");
  }
  
  
  // Create a matrix for the return argument 
  nrow = NCOL;
  ncol = NROW;
  dimsv[0] = ncol;
  dimsv[1] = nrow;
  dimsv[2] = 3;
  plhs[0] = mxCreateNumericArray(3, dimsv,
                                 mxDOUBLE_CLASS, mxREAL);
  if (plhs[0] == NULL)  
  mexErrMsgTxt("Could not create mxArray.\n");
  image = mxGetPr(plhs[0]); 
 
  // fname = mxGetPr(prhs[0]);
  buflen = (mxGetM(prhs[0])*mxGetN(prhs[0]))+1;
  //fname = mxCalloc(buflen, sizeof(char));
  status = mxGetString(prhs[0],  fname, buflen);
  
  /* Do the actual computations in a subroutine */
    hips2image(image, fname);

  //  Load the new matrixes data into plhs[0..2]
  //mxSetPr(plhs[0], imageR);
  // mxSetPr(plhs[1], imageG);
  //mxSetPr(plhs[2], imageB);
  return;
} // end of mexFunction()




//////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////
void hips2image(double* mximage, 
                  char*   fname)
{
  int          i, j, k;
  short int*** image;
 
  image  = new short int ** [NROW];
  for (i=0; i< NROW; i++){ 
      image[i]  = new short int * [NCOL];
      for (j=0; j<NCOL; ++j)
           image[i][j]  = new short int[3];
  }

  read_rgb_image(fname, image);

  for (k=0; k<3; ++k)
  for (j=0; j<NCOL; ++j)
  for (i=0; i<NROW; ++i)  
  mximage[k*NROW*NCOL+j*NROW+i] = image[i][j][k];
 
  delete [] image;

  return;
}




///////////////////////////////////////////////////////////
//
// read_rgb_image(char* fname)
// Aim: 	Read the original colour image: *.RGB.hips
//
// Xiaojuan Feng   
//
///////////////////////////////////////////////////////////
void  read_rgb_image(char* fname, short int*** image)
{
 char* 		fname1=new char[strlen(fname)+13];
 FILE* 		fp;
 unsigned char** red;
 unsigned char** green;
 unsigned char** blue;
 int		i, j;

 strcpy( fname1, fname);
 strcat( fname1, ".col.hips");
 
 fp = fopen(fname1, FILE_READ_BIN);
 if ( !fp ) 
 error1("Error in openning file: *.col.hips !\n");

 // allocation 
 red   = new unsigned char * [NROW];
 green = new unsigned char * [NROW];
 blue  = new unsigned char * [NROW];
 for (i=0; i<NROW; ++i) {
     red[i]   = new unsigned char [NCOL];
     green[i] = new unsigned char [NCOL];
     blue[i]  = new unsigned char [NCOL];
 }

 read_header(fp);
 read_pixels_rgb(fp, red, green,  blue);

 for (i=0; i<NROW; ++i)
 for (j=0; j<NCOL; ++j) {
      image[i][j][0] = red[i][j];
      image[i][j][1] = green[i][j];
      image[i][j][2] = blue[i][j];
 }

 fclose(fp);
 delete [] fname1;
 delete [] red;
 delete [] green;
 delete [] blue;
}  // end of read_rgb_image()



///////////////////////////////////////////////////////////
//
//  read_header(FILE* fp)
//  Aim:	Read the header of a hips image
//
///////////////////////////////////////////////////////////
void	read_header(FILE* fp)
{
char ptr[LINESIZE];

  // Read and toss orig_name.
  get_line(fp, ptr, LINESIZE-1);
  // Read and toss seq_name.
  get_line(fp, ptr, LINESIZE-1);
  // Read num_frame.
  get_line(fp, ptr, LINESIZE-1);
  // Read and toss orig_date.
  get_line(fp, ptr, LINESIZE-1);
  // Read rows.
  get_line(fp, ptr, LINESIZE-1);
  printf("number row=%s \n", ptr);
  // Read cols.
  get_line(fp, ptr, LINESIZE-1);
  printf("number col=%s \n", ptr);
  // Read bits_per_pixel.
  get_line(fp, ptr, LINESIZE-1);
   // Read bit_packing.
  get_line(fp, ptr, LINESIZE-1);
  // Read pixel_format.
  get_line(fp, ptr, LINESIZE-1);
  // Now read and toss lines 
  // until we get one with just a period.
  do{
    get_line(fp, ptr,LINESIZE-1);
  }while (strcmp(ptr, ".\n" ) != 0 );

  return;
} // end of read_header() 




//////////////////////////////////////////////////////////
//
//   get_line( FILE* fp, char* ptr, int size )
//   Aim:	Read a line 
//
//////////////////////////////////////////////////////////
void get_line( FILE* fp, char* ptr, int size )
{
  if ( fgets(ptr,size,fp) == NULL )   
     error1("Error in reading Hips header !\n"); 
 
  return;
} // end of get_line()


///////////////////////////////////////////////////////////
//
//  read_pixels_rgb(FILE *fp)
//  Aim:	Read the pixels of a colour image
//
///////////////////////////////////////////////////////////
void  read_pixels_rgb(FILE *fp, 
                      unsigned char** red,
                      unsigned char** green,
                      unsigned char** blue)
{
 int i, j;

  for (i=0; i<NROW; i++)
  for (j=0; j<NCOL; j++) {
      file_read_bin(&red[i][j],   
                    1*sizeof(unsigned char),1,fp,
                    "Error in read_pixel_rgb !\n");
      file_read_bin(&green[i][j],   
                    1*sizeof(unsigned char),1,fp,
                    "Error in read_pixel_rgb !\n");
      file_read_bin(&blue[i][j],   
                    1*sizeof(unsigned char),1,fp,
                    "Error in read_pixel_rgb !\n");
  }
return;
} // end of read_pixels_rgb()


///////////////////////////////////////////////////////////
//
// file_read_bin(void* ptr,int size,int num,FILE* fp,char* msg)
// Aim:		Read a size ??
//
///////////////////////////////////////////////////////////
void  file_read_bin(void*  ptr,
                    int    size,
                    int    num,
                    FILE*  fp,
                    char*  msg)
{
  //if (fread(ptr,size,num,fp)== NULL) error1(msg);
  if (fread(ptr,size,num,fp)== 0) error1(msg);
  return;
} // end of file_read_bin() 




//////////////////////////////////////////////////////////
//
//  error1(char* msg)
//  Aim:	Print error message and exit 
//
//////////////////////////////////////////////////////////
void  error1(char* msg)
{
  mexPrintf(msg);
  exit(1);
}


