///////////////////////////////////////////////////////////
//
//  	Filename:	label2imageCD7.cc
//	Description:
//	To read label for each pixels into matlab
//	from files: *.lab3, *.seg3.hips
//
//      Copyright 1998 (c)    Xiaojuan Feng
//
///////////////////////////////////////////////////////////
#include  <stdlib.h>
#include  <string.h>
#include  <math.h>
#include  <iostream.h>
#include  <fstream.h>
#include  <math.h>
#include  "mex.h"

#include  "constHips.hh"

#define   NROW		         512
#define   NCOL			 768
#define   FILE_READ_BIN          "rb"
#define   LINESIZE	         5000
#define   MAX_LABELS_OUTPUT      7
#define   MAX_NUMBER_REGIONS     5000
#define   SIZE_OF_DOUBLE_POINTER 4
#define   SIZE_OF_INT_POINTER    4
#define   STOP_CODE              ((unsigned char) 0xff)
#define   LAB_MAGIC              0x30f09f66  
                                 // Magic number in header 
                                 // of labelmap files for 
                                 // unique 
                                 // type-identification
#define LABELMAP_OBJECTS_IDENTITIES	11
                                 // The possible types of 
                                 // labelmap (numbers must
                                 //  start at >1) 
                                 // Hand-labelled object 
                                 // identities

void  error1(char*);
void  file_read_bin(void*,int,int,FILE*,char*);
void  get_line( FILE*, char*, int );
void  read_header(FILE* );
void  read_region(FILE*, short int** );
void  read_segmented_image(char*, short int**);
void  read_labels(FILE*, short int*, int );
int   read_lab3_header(FILE*);
void  read_label_file(char*, short int*);
void  label2imageCD(double*, char*);
short int labels7(short int);


/////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////
void  mexFunction(int            nlhs, 
                  mxArray*       plhs[], 
                  int            nrhs, 
                  const mxArray* prhs[])
{
  double* labels;
  int     nrow, ncol, buflen, status;
  char    fname[30];
  
  // Check for proper number of arguments 
  
  if (nrhs != 1) {
    mexErrMsgTxt("hips2image3d requires one  input argument: *.hips.");
  } else if ( (nlhs < 1)||(nlhs>1)) {
    mexErrMsgTxt("hips2image3d requires one output argument.");
  }
  
  
  // Create a matrix for the return argument 
  nrow = NCOL;
  ncol = NROW;
  plhs[0] = mxCreateDoubleMatrix(nrow, ncol,  mxREAL);
 
  
  // Assign pointers to the various parameters 
  labels = mxGetPr(plhs[0]);
  buflen = (mxGetM(prhs[0])*mxGetN(prhs[0]))+1;
  status = mxGetString(prhs[0],  fname, buflen);

  label2imageCD(labels, fname);
 
  //  Load the new matrixes data into plhs[0..2]
  mxSetPr(plhs[0], labels);

  return;
}




//////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////
void label2imageCD(double* labels_out,  char* fname)
{
  int   i, j;
  short int*  region2label;
  short int** regions;

  
  region2label = new short int[MAX_NUMBER_REGIONS];
  regions = new short int * [NROW];
  for (i=0; i< NROW;i++)  
      regions[i]   = new short int [NCOL];
   
  read_label_file(fname, region2label);
  read_segmented_image(fname, regions);
  
  for (i=0; i<NROW; ++i)
  for (j=0; j<NCOL; ++j)
      labels_out[i*NCOL+j] = region2label[regions[i][j]];
  
  delete [] region2label;
  delete [] regions;

  return;
}


///////////////////////////////////////////////////////////
//
//  read_la(char* fname, short int** region2label)
//  Aim:	Read the label file
//
///////////////////////////////////////////////////////////
void   read_label_file(char* fname, 
                       short int* region2label)
{
 char* 		fname1=new char[strlen(fname)+20];
 FILE 		*fp;
 int             nregs;

 strcpy( fname1, fname);
 strcat( fname1, ".obj.def");
 fp = fopen(fname1, FILE_READ_BIN);
 if ( !fp ) 
 error1("Error in openning file: *.obj.def !\n");
 
 // number of regions composing the image
 nregs = read_lab3_header(fp);
 if (nregs > MAX_NUMBER_REGIONS) {
 mexPrintf("The number (%d) of regions is too large.\n");
 exit(1);
 }

 // map of the labels
 read_labels(fp, region2label, nregs);
 
 fclose(fp);
 delete [] fname1;
} // end of read_label_file()


///////////////////////////////////////////////////////////
//
//  read_lab3_header(FILE *fp)
//  Aim:	Read the header of a label file
//
///////////////////////////////////////////////////////////
int   read_lab3_header(FILE *fp)
{
 int 	status, validity, tmp_nregions;

  file_read_bin(&status, sizeof(int), 1, fp,
               "Label File Read Failed ! \n");
  if (status != LAB_MAGIC)
     error1("Label File Read Failed !\n");
  file_read_bin(&validity, sizeof(int), 1, fp,
               "Label File - Bad Format !\n");

  //For backward compatibility check validity flag exists:
  if ( (validity==0)||(validity==1) ) {
     //read number of regions in this labelmap
     file_read_bin(&tmp_nregions, sizeof(int), 1, fp,
                  "Label File - Bad Format !\n");
  }
  else{
     //For backward compatibility with v1.0 label files:
     tmp_nregions = validity;
     validity = -1;
  }
  if (tmp_nregions<=0) error1("Label File - Bad Format\n");
  file_read_bin(&status, sizeof(int), 1, fp,
                "Label File - Bad Format\n");

  #ifdef DEBUG
  if (DEBUGLEVEL > 2) {
     fprintf(stderr, "Loading labelmap file...\n");
     fprintf(stderr, "Label type:");
  }
  #endif

  switch(status){
        case LABELMAP_OBJECTS_IDENTITIES:
             #ifdef DEBUG
             if (DEBUGLEVEL > 2)
                fprintf(stderr, "Object identities\n");
             #endif
             break;

        default: error1("Unknown\n");
                 break;
            
        #ifdef DEBUG
        if(DEBUGLEVEL > 2)  fprintf(stderr, "Unknown\n");
        #endif
  }

  return(tmp_nregions);
} // end of read_lab3_header()


///////////////////////////////////////////////////////////
//
//  Hips::read_labels(FILE *fp, short int** region2label)
//  Aim:	Read the label fro each pixel
//
///////////////////////////////////////////////////////////
void  read_labels(FILE *fp, 
                  short int* region2label,
                  int nregs)
{
int 	i;
int 	status,  nitems;
int     serialidx;
int*    serialidxlist;
double  labelid;
double  qualifier;
double* labelidlist;

 for(i=0; i<nregs; i++) {
    file_read_bin(&labelid, sizeof(double),1,fp,
	          "Label File - Bad Format \n");
    file_read_bin(&labelidlist,
                  SIZE_OF_DOUBLE_POINTER,1,fp,
		  "Label File - Bad Format \n");
    file_read_bin(&serialidx,sizeof(int),1,fp,
	          "Label File - Bad Format \n");
    region2label[i] = labels7(serialidx);


    file_read_bin(&serialidxlist,
                  SIZE_OF_INT_POINTER,1,fp,
		  "Label File - Bad Format \n");
    file_read_bin(&qualifier,sizeof(double),1,fp,
	      	  "Label File - Bad Format \n");
    nitems = 0;
    if (labelidlist != (double *) NULL) {
    //This regions is multiply labelled:
    do{
      //Read label and its serial index:
      file_read_bin(&labelid,sizeof(double),1,fp,
                    "Label File - Bad Format \n");
      status = (int)(((*(unsigned char*)&labelid) 
                       & STOP_CODE)
                      !=STOP_CODE);
      //Status = 1 <=> left != right;
      //Status = 0 <=> left  = right;
      //cout<<"labelid: "<<status<<endl;
      if(status != 0){
	//not at the end of higher order labels
	nitems++;
	file_read_bin(&serialidx,sizeof(int), 1,fp,
		      "Label File - Bad Format \n");
        region2label[i] = labels7(serialidx);

	//Read label qualifier (no serial index):
      	file_read_bin(&labelid,sizeof(double), 1,fp,
		      "Label File - Bad Format \n");
      }
    }while (status!=0);
    labelidlist    = (double *) (NULL);
    serialidxlist  = (int *) (NULL);
    }
 }
  return ;
} // end of read_labels()



///////////////////////////////////////////////////////////
//
//  Hips::read_segmented_image(char* fname)
//  Aim:	Read the segmented image
//
///////////////////////////////////////////////////////////
void    read_segmented_image(char*       fname, 
                             short int** regions)
{
char* 		fname1=new char[strlen(fname)+20];

FILE*		fp;

 strcpy( fname1, fname);
 strcat( fname1, ".obj");
 fp = fopen(fname1, FILE_READ_BIN);
 if ( !fp ) 
 error1("Error in opening file: *.seg3.hips !\n");
 
 //Reading and tossing the header of the *.seg3.hips file
 read_header(fp);

 //Reading the regions in the *.seg3.hips file
 read_region(fp, regions);
 
 fclose(fp);
 delete [] fname1;
} // end of read_label_image()
 


///////////////////////////////////////////////////////////
//
//  read_header(FILE* fp)
//  Aim:	Read the header of a hips image
//
///////////////////////////////////////////////////////////
void	read_header(FILE* fp)
{
char ptr[LINESIZE];

  // Read and toss orig_name.
  get_line(fp, ptr, LINESIZE-1);
  // Read and toss seq_name.
  get_line(fp, ptr, LINESIZE-1);
  // Read num_frame.
  get_line(fp, ptr, LINESIZE-1);
  // Read and toss orig_date.
  get_line(fp, ptr, LINESIZE-1);
  // Read rows.
  get_line(fp, ptr, LINESIZE-1);
  // Read cols.
  get_line(fp, ptr, LINESIZE-1);
  // Read bits_per_pixel.
  get_line(fp, ptr, LINESIZE-1);
   // Read bit_packing.
  get_line(fp, ptr, LINESIZE-1);
  // Read pixel_format.
  get_line(fp, ptr, LINESIZE-1);
  // Now read and toss lines 
  // until we get one with just a period.
  do{
    get_line(fp, ptr,LINESIZE-1);
  }while (strcmp(ptr, ".\n" ) != 0 );

  return;
} // end of read_header() 




///////////////////////////////////////////////////////////
//
//   get_line( FILE* fp, char* ptr, int size )
//   Aim:	Read a line 
//
///////////////////////////////////////////////////////////
void get_line( FILE* fp, char* ptr, int size )
{
  if ( fgets(ptr,size,fp) == NULL )   
     error1("Error in reading Hips header !\n");

  return;
} // end of get_line()




///////////////////////////////////////////////////////////
// 
//  Hips::read_region(FILE* fp)
//  Aim:	Read region ??
//
///////////////////////////////////////////////////////////
void  read_region(FILE* fp, short int ** regions)
{
int 		i;
short int   *pix;

  pix = new short int[NCOL+1];               
  //The first row is the border.
  file_read_bin(pix, (NCOL+1)*sizeof(short int), 1, fp,
  "Error in reading the first row of ssegmented image!\n");
  delete [] pix ;

  // For each row of the image (the last is tossed)
  pix = new short int[2];
  for (i=0; i<NROW; i++) {
      //mexPrintf("%d\n", i);
      // The first two pixels belong to the border 
      // of the previous and the actual row
      file_read_bin(pix, 2*sizeof(short int),1,fp, 
                    "Error in read_region 1 !\n");
      file_read_bin(regions[i], NCOL*sizeof(short int),
                    1,fp,
                    "Error in read_region 2 !\n");
  }
  delete [] pix ;

  return;
} // end of read_region () 



//////////////////////////////////////////////////////////
//
// Hips::file_read_bin(void* ptr,int size,int num,FILE* fp,char* msg)
// Aim:		Read a size ??
//
///////////////////////////////////////////////////////////
void file_read_bin(void* ptr,
                   int   size,
                   int   num,
                   FILE* fp,
                   char* msg)
{
  if (fread(ptr,size,num,fp)== 0) error1(msg);
  return;
} // end of file_read_bin() 





///////////////////////////////////////////////////////
//
//  error1(char* msg)
//  Aim:	Print error message and exit 
//
///////////////////////////////////////////////////////
void  error1(char* msg)
{
  mexPrintf(msg);
  exit(1);
}



///////////////////////////////////////////////////////////
//
// Hips::labels7 (short int serialidx)
// Aim:		??
//
///////////////////////////////////////////////////////////
short int   labels7(short int serialidx)
{
short int res;

  switch(serialidx){
    case Atmospheric_Phenomena:
    case AP_Clear_Sky:
    case AP_Cloud:
    case AP_Mist_Smoke:
    case AP_Other:
      res = CLOUDS ;
      break;
    case Landscape:
    case L_Stone_Rock:
    case L_Soil:
    case L_Vegetation:
    case L_V_Bush:
    case L_V_Grass:
    case L_V_Hedge:
    case L_V_Tree:
    case L_V_Other:
    case L_Water:
    case L_Other:
    case RO_RB_Grass:
    case RO_RB_Leaves:
    case RO_RS_Leaves:
    case RO_RS_Grass:
      res = VEGETATION ;
      break;
    case RO_RS_Marking:
    case RO_RS_M_Centre:
    case RO_RS_M_Edge:
    case RO_RS_M_Other:
      res = ROAD_SURFACE_MARKING ;
      break;
    case RO_Road_Surface:
    case RO_RS_Concrete:
    case RO_RS_Drain:
    case RO_RS_Gravel:
    case RO_RS_Mud:
    case RO_RS_Stone_Rock:
    case RO_RS_Tarmac:
    case RO_RS_Water:
    case RO_RS_Other:
    case Road_Object:
    case RO_Road_Border:
    case RO_RB_Gravel:
    case RO_RB_Mud:
    case RO_RB_Stone_Rock:
    case RO_RB_Tarmac:
    case RO_RB_Concrete:
    case RO_RB_Drain:
    case RO_RB_Kerb_Stone:
    case RO_RB_Water:
    case RO_RB_Other:
      res = ROAD_SURFACE_BORDER ;
      break;
    case Building:
    case B_Chimney:
    case B_Door:
    case B_Roof:
    case B_Wall:
    case B_Window:
    case B_Other:
    case RO_Bridge:
    case BO_Gate:
    case BO_Wall:
      res = BUILDING ;
      break;
    case BO_Fence:
    case RO_Road_Sign:
    case Electrical_Object:
    case EO_Electricity_Pylon:
    case EO_Lamp_Post:
    case EO_Telegraph_Pole:
    case EO_Wire:
    case EO_Other:
    case RO_Traffic_Signal:
    case RO_Other:
      res = STREET_FURNITURE ;
      break;
    case Mobile_Object:
    case MO_Bike:
    case MO_B_Body:
    case MO_B_Wheel:
    case MO_B_Other:
    case MO_Car:
    case MO_C_Body:
    case MO_C_Wheel:
    case MO_C_Window:
    case MO_C_Other:
    case MO_Person:
    case MO_Truck:
    case MO_T_Body:
    case MO_T_Wheel:
    case MO_T_Window:
    case MO_T_Other:
    case MO_Van:
    case MO_V_Body:
    case MO_V_Wheel:
    case MO_V_Window:
    case MO_V_Other:
    case MO_Other:
      res = MOBILE_OBJECT_CAR ;
      break;
    case Undefined_Other:  
      res = UNDEFINED;
      break;
    default:
      res = UNDEFINED;
      break;
  }

return(res);
} // end of labels11()

