/////////////////////////////////////////////////////////////
//
//  filename:   hipsCD.hh
//  class Node:     Hips
//
//
// Copyright (c) Xiaojuan Feng 1998
////////////////////////////////////////////////////////////

#ifndef 	HIPS_H
#define         HIPS_H

#include <stdio.h>
#include "constHips.hh"

// Define the  structures
typedef struct {
    char* orig_name;	// An indication of the originator 
                        // of this sequence. 
    char* seq_name;	// The sequence name. 
    int   num_frame;	// The number of frames in this 
                        // sequence. 
    char* orig_date;	// The date the sequence was 
                        // originated.
    int   nrow;	 	// The number of rows in each image, 
    			// the height.
    int   ncol;  	// The number of columns in 
                        // each image, the width.
    int  bits_per_pixel;// The number of significant bits 
                        // per pixel.
    int  bit_packing;   // Nonzero if the bits were packed 
                        // such as to eliminate any unused 
                        // bits eliminate any unused bits 
                        // resulting from a bits_per_pixel 
                        // value which was not an 
                        // even multiple of eight.
    int   pixel_format; // An indication of the format of
                        // each pixel.
    char* seq_history;	// A description of the sequence of
                        // transformations leading up to the 
                        // current image.
    char* seq_desc;	// A free form description of the 
                        // contents of the sequence.
  }  HipsHeader;

 // Rgb
 typedef struct rgb1{
    un_char**  red;
    un_char**  green;
    un_char**  blue;
 } Rgb;

 // LabelMape
 typedef struct  {
    double labelid;        // Unique numeric primary label 
                           // Id (bit 0 = lock bit)
    double qualifier;	   // Qualifier for primary label 
    double *labelidlist;   // STOP-terminated list of Id's 
                           // for multiple labels 
    int    serialidx;	   // Serial index of label in array 
                           // of label strings 
                           // (the label for the region) 
    int    *serialidxlist; // STOP-termin. list of serial 
                           // indexes  (multiple labs)
  } LabelMape;



// Hips
class  Hips {

public:
        Hips(char*);
        Hips();
        ~Hips();
  void  GetVE(char*, int); 
  void  ViewOutput(char*);
  int   GetNumRow();
  int   GetNumCol();
  void  DeleteAllocation();
  void  GetNumClass(FILE**);
  void  OutputRandomPixel(int,unsigned int,FILE**);
  void  ReduceLabelImage(s_int, FILE**);
  void  ReduceFuzzyImage(s_int, FILE**);
  void  CalculateFeature(s_int, FILE**);
  void  CalculateNormFeature(float**,s_int,FILE**);
  void  CalculateSelectedNormFeature(float**,s_int,FILE**);
  un_char*  GetAPixelValue(int, int);

private:
	HipsHeader   	hips_header, seg_hips_header;
	Rgb          	rgb;
  	LabelMape    	*idl_regtbl;
        s_int**  	region;
        int    		nregs;
        int    		num_class[MAX_LABELS_OUTPUT+1];
        int             num_row, num_col;


  void  ReadRGBImage(char*);// Read the file: 
                                    // *.RGB.hips
  void	ReadHeader(FILE*, HipsHeader*);      
                                    // Read the header of 
                                    // the *.RGB.hips
  void  GetLine(FILE*, char*, int);
  void  AllocateRGB(Rgb*);
  void  ReadPixelRGB(FILE*);


  void  ReadLabelFile(char*);// Read the file: *.labs
  int   ReadLab3Header(FILE*);
  void  ReadLabel(FILE*);
  void  Colour11(un_char*, un_char*,un_char*,int);
  void  DeleteAllocatedRGB(Rgb*);


  void  ReadSegmentedImage(char*);// Read the file: 
                                          // *.seg3.hips
  void  ReadRegion(FILE*);


  void  hips2ppm(char*, Rgb); // Write the hips 
                                        // image into 
                                        // the ppm format and
                                        // display *.ppm file
  void  Writeppm(FILE*, Rgb, int);
  void  ppmHeader(FILE*, int);
 
  void  hips2pgm(char* fname);
  void  pgmHeader(FILE*, int);
  s_int Label7(s_int);

  void  seg2ppm(char*);


  void  FileReadBin(void*,int,int,FILE*,char*);
  void  Error1(char*);
  void  ViewImage(char*);


}; // end of class Hips


#endif









