/////////////////////////////////////////////////////////////
//  class Node:     node.hh
//
//  Xiaojuan Feng
//  19th June, 1997
//
//  Modifications:
//        	07/11/97 -- Training from label-images
////////////////////////////////////////////////////////////

#ifndef 	NODE_H
#define         NODE_H

#include	"const.hh"

///////////////////////////////////////////////////////////
//  Struct:
//                   ^   |
//  lambda_message   |   V  pi_message
//                 --------
//                 |      |
//                 | node | 
//                 |      |
//                 -------- 
//////////////////////////////////////////////////////////

typedef struct {
  int 	        parent; 
  int 	        child1; // index for the first child
  float		lambda_message[NUMBER_CLASSES];
  float         pi_message[NUMBER_CLASSES];

} Node;




///////////////////////////////////////////////////////////
//  Aim: To train  from label-images
//  Struct:
//                   ^   |
//  lambda_message   |   V  pi_message
//                 --------
//                 |      |
//                 | node | 
//                 |      |
//                 --------
// Xiaojuan Feng  01/11/97 
//////////////////////////////////////////////////////////
typedef struct {
  int 	        parent; 
  int 	        child1; // index for the first child
  float		lambda_value[NUMBER_CLASSES];
  float		lambda_message[NUMBER_CLASSES];
  float         pi_value[NUMBER_CLASSES];
  float         pi_message[NUMBER_CLASSES];
} NodeTrain;




///////////////////////////////////////////////////////////
//  Aim: To train  from label-images
//       with complex CPTs.
//  Struct:
//                   ^   |
//  lambda_message   |   V  pi_message
//                 --------
//                 |      |
//                 | node | 
//                 |      |
//                 --------
//  CPT: From parent to this node
// Xiaojuan Feng  01/11/97 
// 	09/01/98
//////////////////////////////////////////////////////////
typedef struct {
  int 	        parent; 
  int 	        child1; // index for the first child
  int		condp_index; //
  float		lambda_value[NUMBER_CLASSES];
  float		lambda_message[NUMBER_CLASSES];
  float         pi_value[NUMBER_CLASSES];
  float         pi_message[NUMBER_CLASSES];
} NodeTrain2;




///////////////////////////////////////////////////////////
//  Aim: To train  from label-images
//       with complex CPTs.
//  Struct:
//                   ^   |
//  lambda_message   |   V  pi_message
//                 --------
//                 |      |
//                 | node | 
//                 |      |
//                 --------
//  CPT: From parent to this node
// Xiaojuan Feng  12/01/99 
//
//////////////////////////////////////////////////////////
typedef struct {
  int 	        parent; 
  int 	        child1; // index for the first child
  int		condp_index; //
  float		lambda_value[NUMBER_CLASSES];
  float		lambda_message[NUMBER_CLASSES];
  float         pi_value[NUMBER_CLASSES];
  float         pi_message[NUMBER_CLASSES];
  float         prior[NUMBER_CLASSES];
  double        log_scale; // log scaling factor
  double        log_marg_prob;// log marginal probability
} NodeTrain3;


#endif









