/////////////////////////////////////////////////////////////
//
//  Filename:	 tsbn.hh
//  Description: Definition file for the TSBN class.
//
//  Copyright (c) 1998 Xiaojuan Feng
//
////////////////////////////////////////////////////////////

#ifndef		TSBN_h
#define		TSBN_h

#include	"const.hh"
#include	"node.hh"


class TSBN {

public:
  /* 
        TSBN(s_int	nl, 
             s_int*	nrows, 
             s_int*	ncols,
             s_int*	np,    
             s_int*	nc, 
             s_int	nv, 
             float**    ve, 
             float*     prob);
        TSBN(s_int	nl, 
	     s_int*	nrows, 
             s_int*	ncols,
             s_int*	np,  
             s_int*	nc, 
             s_int	nv,
             s_int*	labels,
             float*     priors,
             float***   probs );  
	     */
        TSBN(s_int  nl,
             s_int* nrows,
             s_int* ncols,
             s_int  ncpt,
             s_int* np,
             s_int* nc,
             s_int  nv);
                       // Training from label-image
       TSBN(s_int  nl, 
            s_int* nrows, 
           s_int* ncols,
           s_int  ncpt,
           s_int* np, 
           s_int* nc, 
           s_int  nv,
           char*  type );
        ~TSBN();
   // set the conditional probabilities
  //void		SetCPT(float** theta);
   void		SetCPT1(float*** probs);
   void		SetCPT6(float*** probs);
	  
   void		SetEvidence( s_int* labels);
   void 	SetVirtualEvidence( float** labels);
   void 	SetInitialValue( );

   void 	InitialLambdaPi();
   void 	PostRootPiMessage();
   void 	PostRootPiMessageBest();
   void         PostLeafLambdaMessage(); 
   void         PostLeafLambdaMessage2();  
   void		PostLeafLambdaMessageBest();   
   void 	Output(char* fname);
   void 	Output(s_int* labels1, float** conf);
   void      	Output(double* labelB);
   void 	Output_prior(char* fname);
   void 	OutputEntropy(char* fname);
   void 	PropagateB();
   void 	PropagateBBest();
   void 	PropagateC();
   void 	PropagateCBest();
   void         PassMessage();
   void         PassMessageBest();
   void 	EstimateCPT( float*** probs_out );
   void 	EstimateCPTConstrained(
                         float*** probs_out);
   void 	MatchLabel(float*** probs_out);
   float 	Loglikelihood(  );
   void         Belief2VE(  );
   void         CalculatePrior( s_int type );
   void         CalculateLambdaMargProb(int node);
   void         CalculateLambdaMargProb(); 
   float*       OutputLogMargProb();
public:

   s_int	num_levels; 	 // number of levels
   s_int        num_cpts; 	 // number of CPTs
   s_int*       num_rows;        // numbers of rows at 
                                 // different levels
   s_int*       num_cols;        // numbers of columns 
                                 // at different levels 
   s_int*	num_parents; 	 // numbers of parents 
                                 // for different levels
   s_int*       num_children; 	 // numbers of children 
                                 // for different levels
   s_int 	num_classes; 	 // number of possible 
                                 // classes

private:
   s_int        evidence_type;   
   int	        num_nodes; 	 // number of nodes
   int	        nitemlambda; 	 // number of nodes
				 // in stock: rlambda
   int	        nitempi;	 // number of nodes in 
                                 // stock: rlambda
   int 	        nitemstack;      // the maximum number
                                 //  of stack elements
   int*	        rlambda;	 // stack for nodes receiving
				 // new lambda messages
   int*	        rpi;		 // stack for nodes receiving
				 // new pi  messages
   int*         nodes_levels;    // number of nodes for
                                 //  each level
   int*         sum_nodes_levels;// total number of nodes 
                                 // before the current level
   float***   	condp;           // (index, parent, child, )
                                 //  condp[index][i][j]
                                 // 10/01/98
   float**      virtual_evidence;// A num_node * num_classes
                                 //  matrix
   float*       prior_prob;      // the prior probilities
                                 // at root node
   float        scale_d;         // used in scaling process
   s_int*       hard_evidence;   // A num_node matrix
  NodeTrain3* 	tree;  // Temp:  13/01/99 Think deeply!!!
  //NodeTrain2* 	tree1;  // Temp:  13/01/99 Think deeply!!!
                    
   s_int        GetBestClass(int node);
   s_int        GetANodeLevel(int node);
   int          GetNumNode();
   void         GetNodeLevel();
   float        GetCP(int   child_node,
                      s_int index_parent,
                      s_int index_child);
   void         GetLambdaValue(int node);
   float*       GetLambdaValueUnscaled(int node);
   void  	GetPiValue(int node);
   void  	GetPiValueBest(int node);
   float*       GetBEL(int node);
   float  	CalEntropy(int node) ;
   void         ScaleArray( float* array, float scale );
   void         EstablishRelation();
   void         UpdateLambdaMessage(int node);
   void         UpdateLambdaMessageBest(int node);
   void         UpdatePiMessage(int node);
   void         Push(int* stock, int* nitem, int node);
   int          Pop(int* stock, int* ntiem);

}; // end of class TSBN

#endif
