/////////////////////////////////////////////////////////////
//
// Filename:	classifyTSBN.cc
// Description:	Implementation file for ClassifyTSBN class.
//
// Copyright (c) 1998   Xiaojuan Feng
//
/////////////////////////////////////////////////////////////



#include <iostream.h>
#include <fstream.h>
#include <limits.h>
#include <stdlib.h>

#include "classifyTSBN.hh"

////////////////////////////////////////////////////////////
//
//
////////////////////////////////////////////////////////////
ClassifyTSBN::ClassifyTSBN(s_int  nclass, 
                           s_int  nlevel,
                           s_int  nrow,   
                           s_int  ncol,
                           s_int  ncpt,   
                           s_int  nimage)
            : InputTSBN(nclass, 
                        nlevel,
                        nrow,   
                        ncol,
                        ncpt,   
                        nimage)
{
} // end of: ClassifyTSBN::ClassifyTSBN

////////////////////////////////////////////////////////////
//
//
////////////////////////////////////////////////////////////
ClassifyTSBN::~ClassifyTSBN()
   {
   }




////////////////////////////////////////////////////////////
//
//  void  ClassifyTSBN::Classify(float** images, 
//                               float*** probs, 
//                               char* fname)
//
////////////////////////////////////////////////////////////
void  ClassifyTSBN::Classify( float** images, 
                              float*** probs, 
                              char* fname)
{

 float**        labels = ChangeFormat( images );
 delete images;

 // set the priors an d CPT to b.net
 //tree->SetCPT11( probs );
 tree->SetCPT6( probs );
 tree->SetInitialValue();
 tree->SetVirtualEvidence(labels);

 //mexPrintf("1:\n");
 tree->PostLeafLambdaMessage(); 
 tree->PropagateB();
 // post pi-message
 //mexPrintf("2:\n");
 tree->PostRootPiMessage(); 
 tree->PropagateC(); 

 tree->Output(fname);
 
} // end      

////////////////////////////////////////////////////////////
// void  ClassifyTSBN::Classify( char* fimage1, 
//                               char* fimage2, 
//                               float*** probs,   
//                               char* fname)
//
////////////////////////////////////////////////////////////
void  ClassifyTSBN::Classify( char* fimage1, 
                              char* fimage2, 
                              float*** probs,   
                              char* fname)
{
 s_int      i, j, k;
 float**    conf = fmatrix(8,8);
 s_int      nrow = tree->num_rows[tree->num_levels-1];
 s_int      ncol = tree->num_cols[tree->num_levels-1];
 s_int      nclass = tree->num_classes;
 int	    nsize = nrow*ncol;
 s_int*     images1 = new s_int [nsize];
 s_int*     labels1 = new s_int [nsize]; 
 float**    images2 = fmatrix(nsize, (int)nclass);
 float**    labels2 = fmatrix(nsize, nclass);
 float      temp;
 FILE           *fp1, *fp2;
   
 
 // set the priors an d CPT to b.net
 //tree->SetCPT1( probs );
 tree->SetCPT6( probs );
 //tree->set_cond_prob_3( probs );

 if ((fp1=fopen( fimage1, "r")) == NULL)
    errors("file_hard doesn't exist\n"); 
 if ((fp2=fopen( fimage2, "r")) == NULL)
    errors("file_ve doesn't exist\n"); 

 
 for (i=0; i<num_image; ++i) {
     
     // input
     for (j=0; j<nsize; ++j) {
          fscanf(fp1, "%f ", &temp);
          images1[j] = temp;
          for (k=0; k<nclass; ++k) 
              fscanf(fp2, "%f ", &images2[j][k]);
     }

     // if ( i!=13 ) continue;
     printf("image: %d\n", i);

 
     ChangeFormat( images2, labels2 );
     ChangeFormat( images1, labels1 );
 
     tree->SetInitialValue();
     tree->SetVirtualEvidence( labels2 );
     tree->PassMessage();

     tree->Output( labels1, conf);
 }
 
 fclose(fp1);
 fclose(fp2);

 ofstream out(fname); 
 for (i=0; i<8; ++i) {
     for (j=1; j<8; ++j)
         out << conf[i][j] << " ";
     out << "\n";
 }
 out.close();

 delete [] images2 ;
 delete [] images1 ;
 delete [] labels1;
 delete [] labels2; 
} // end      

     


////////////////////////////////////////////////////////////
//void  ClassifyTSBN::ClassifyBest( char* fimage1, 
//                                  char* fimage2, 
//                                  float*** probs,   
//                                  char* fname)
////////////////////////////////////////////////////////////
void  ClassifyTSBN::ClassifyBest( char* fimage1, 
                                  char* fimage2, 
                                  float*** probs,   
                                  char* fname)
{
 s_int      i, j, k;
 float**    conf  = fmatrix(8,8);
 s_int      nrow = tree->num_rows[tree->num_levels-1];
 s_int      ncol   = tree->num_cols[tree->num_levels-1];
 s_int      nclass = tree->num_classes;
 int	    nsize = nrow*ncol;
 s_int*     images1 = new s_int [nsize];
 s_int*     labels1 = new s_int [nsize]; 
 float**    images2 = fmatrix(nsize, (int)nclass);
 float**    labels2 = fmatrix(nsize, nclass);
 float      temp;
 FILE       *fp1, *fp2;
   
 
 // set the priors and CPT to b.net
 //tree->SetCPT1( probs );
 tree->SetCPT6( probs );
 //tree->set_cond_prob_3( probs );

 if ((fp1=fopen( fimage1, "r")) == NULL)
     errors("file_hard doesn't exist\n"); 
 if ((fp2=fopen( fimage2, "r")) == NULL)
     errors("file_ve doesn't exist\n"); 

 for (i=0; i<num_image; ++i) {
     printf("image: %d\n", i);
     
     // input
     for (j=0; j<nsize; ++j) {
          fscanf(fp1, "%f ", &temp);
          images1[j] = temp;
          for (k=0; k<nclass; ++k) 
              fscanf(fp2, "%f ", &images2[j][k]);
     }

     ChangeFormat( images2, labels2 );
     ChangeFormat( images1, labels1 );
 
     tree->SetInitialValue();
     tree->SetVirtualEvidence( labels2 );
     tree->PassMessageBest();

     tree->Output( labels1, conf);
     // tree->output( "f");
  
 }
 
 fclose(fp1);
 fclose(fp2);

 ofstream out(fname); 
 for (i=0; i<=nclass; ++i) {
     for (j=1; j<=nclass; ++j)
         out << conf[i][j] << " ";
     out << "\n";
 }
 out.close();

 delete [] images2 ;
 delete [] images1 ;
 delete [] labels1;
 delete [] labels2; 
} // end    




 


////////////////////////////////////////////////////////////
//void  ClassifyTSBN::Classify( double* ve,  
//                              float*** probs, 
//                              double* labelB )
//////////////////////////////////////////////////////////// 

void  ClassifyTSBN::Classify( double* ve,  
                              float*** probs, 
                              double* labelB )
{
 s_int      j, k;
 s_int      nrow = tree->num_rows[tree->num_levels-1];
 s_int      ncol = tree->num_cols[tree->num_levels-1];
 s_int      nclass = tree->num_classes;
 int	    nsize = nrow*ncol;
 float**    images2 = fmatrix(nsize, (int)nclass);
 float**    labels2 = fmatrix(nsize, nclass);

 // set the priors an d CPT to b.net
 //tree->SetCPT1( probs );
 tree->SetCPT6( probs );
 //tree->set_cond_prob_3( probs );



 // input
 for (j=0; j<nsize; ++j) 
 for (k=0; k<nclass; ++k) 
      images2[j][k] = ve[k*nsize+j];

 ChangeFormat( images2, labels2 );
 
 printf("    bottom-up propagating .....\n");
 //flush;
 tree->SetInitialValue();
 tree->SetVirtualEvidence( labels2 );
 tree->PostLeafLambdaMessage(); 
 tree->PropagateB();

 printf("    top-down propagating .....\n");
 //flush;
 tree->PostRootPiMessage(); 
 tree->PropagateC(); 
	

 printf("    classifying .....\n");
 //flush;
 tree->Output( labelB);

 delete [] images2 ;
 delete [] labels2; 
} // end      

     
 




 ////////////////////////////////////////////////////////////
 //  void  TrainBeliefNet::classify( float*** probs, 
 //                                  float tol, s_int maxiter)
 //
 //  Function: To classify an image using belief 
 //            net to combine the priros
 //            and the scaled likelihood.   
 //  Modification:
 //  		15/02/98
 ////////////////////////////////////////////////////////////
 void  ClassifyTSBN::ClassifyBest( double*  ve,  
                                   float*** probs, 
                                   double* labelB )
 {
 s_int      j, k;
 s_int      nrow = tree->num_rows[tree->num_levels-1];
 s_int      ncol = tree->num_cols[tree->num_levels-1];
 s_int      nclass = tree->num_classes;
 int	    nsize = nrow*ncol;
 float**    images2 = fmatrix(nsize, (int)nclass);
 float**    labels2 = fmatrix(nsize, nclass);
			   
 // set the priors an d CPT to b.net
 //tree->SetCPT1( probs );
 tree->SetCPT6( probs );
 //tree->set_cond_prob_3( probs );

 // input
 for (j=0; j<nsize; ++j) 
 for (k=0; k<nclass; ++k) 
     images2[j][k] = ve[k*nsize+j];
 
 ChangeFormat( images2, labels2 );
 
 printf("    bottom-up propagating .....\n");
 //flush;
 tree->SetInitialValue();
 tree->SetVirtualEvidence( labels2 );
 tree->PostLeafLambdaMessageBest(); 
 tree->PropagateBBest();

 printf("    top-down propagating .....\n");
 //flush;
 tree->PostRootPiMessageBest(); 
 tree->PropagateCBest(); 
	
 printf("    classifying .....\n");
 //flush;
 tree->Output( labelB);
 tree->OutputEntropy( "entropy.dat");

 delete [] images2 ;
 delete [] labels2; 
} // end      

     
 
///////////////////////////////////////////////////////////
//
// 28/08/98
///////////////////////////////////////////////////////////

void ClassifyTSBN::ClassifyBestCondp(char* fimage1,
                                     char* fimage2,
                                     float*** probs,
                                     char* fcondp)
{
  int 	    i, j,k,pi,pj,pk;
  s_int     nlevel = tree->num_levels;
  s_int     nrow = tree->num_rows[nlevel-1];
  s_int     ncol = tree->num_cols[nlevel-1];
  s_int     ncpt = tree->num_cpts;
  s_int     nclass = tree->num_classes;
  float     temp;
  int	    nsize = nrow*ncol;
  s_int*    images1 = new s_int [nsize];
  s_int*    labels1 = new s_int    [nsize]; 
  float*    fval = new float[num_image];
  float*    fval1 = new float[num_image];
  float*    sval = new float[num_image];  
                      // P(x*|y)=\prod p(x*_i|y*_i) 
  float**   images2 = fmatrix(nsize, (int)nclass);
  float**   labels2 = fmatrix(nsize, nclass);
  float*    compen = new float[(int)nclass];
  FILE          *fp1, *fp2;

  // set the priors and CPT to b.net
  //tree->SetCPT1( probs ); // One CPT on each level
  tree->SetCPT6( probs ); // six CPT on level 1
  
  //  obain the evidences
  if ((fp1=fopen( fimage1, "r")) == NULL)
     errors("file_hard doesn't exist\n"); 
  if ((fp2=fopen( fimage2, "r")) == NULL)
     errors("file_ve doesn't exist\n"); 

  for (i=0; i<num_image; ++i) {
      printf("image: %d\n", i);
      // input
      fval[i] = 0; 
      sval[i] = 0;
      for (j=0; j<nsize; ++j) {
          fscanf(fp1, "%f ", &temp);
          images1[j] = temp;
          for (k=0; k<nclass; ++k) 
              fscanf(fp2, "%f ", &images2[j][k]);

          if ((images1[j]>=0)&&
              (images2[j][images1[j]]>ZERO)) {
              fval[i] +=
              (float)log((double)images2[j][images1[j]]);

	      //  This part should be changed when this 
              // function is called.  Xiaojuan  10/01/99



	      // raw MLP prediction
              //sval[i] +=
              // (float)log((double)(images2[j][images1[j]]*
              //         prior_train[images1[j]]));

              // compensation
              // normalise the compensated prediction
              //temp = 0.0;
              //for (k=0; k<nclass; ++k) {
              //    compen[k] = images2[j][k]*prior_true[k];
              //    temp += compen[k];
              //}
              //sval[i] += 
              //(float)log((double)(compen[images1[j]]/temp));
          }
          else 
            for (k=0; k<nclass; ++k) images2[j][k] = 0.0;
       }

      ChangeFormat( images2, labels2 );
      ChangeFormat( images1, labels1 );
      // set initial value for scale_d 
      tree->SetInitialValue();
      // set hard evidence
      tree->SetEvidence( labels1 );
      // message passing once
      tree->PassMessage();
      //  the loglikelihood
      fval1[i] = tree->Loglikelihood();

      // set initial value for scale_d 
      tree->SetInitialValue();
      // set virtual evidence
      tree->SetVirtualEvidence( labels2 );
      // message passing once
      tree->PassMessage();
      //  the loglikelihood
      fval1[i] -= tree->Loglikelihood();

  }

  fclose(fp1);
  fclose(fp2);

  // output 
  ofstream out(fcondp); 
  for (i=0; i<num_image; ++i) {
    out <<fval[i]+fval1[i] <<" "<<sval[i];
    out << "\n";
  }
  out.close();

 } // end of iteration_CML_scg


