/////////////////////////////////////////////////////////////
//
// Filename:	classifyTSBN.cc
// Description:	Implementation file for CodeTSBN class.
//
// Copyright (c) 1999   Xiaojuan Feng
//
/////////////////////////////////////////////////////////////



#include <iostream.h>
#include <fstream.h>
#include <limits.h>
#include <stdlib.h>

#include "codeTSBN.hh"

////////////////////////////////////////////////////////////
//
//
////////////////////////////////////////////////////////////
CodeTSBN::CodeTSBN(s_int  nclass, 
                   s_int  nlevel,
                   s_int  nrow,   
                   s_int  ncol,
                   s_int  ncpt,   
                   s_int  nimage)
            : InputTSBN(nclass, 
                        nlevel,
                        nrow,   
                        ncol,
                        ncpt,   
                        nimage,
                       (s_int)1)
{

 s_int*  nparents  = SetNumParent( nlevel );
 s_int*  nchildren = SetNumChildren( nrow, ncol, nlevel );
 s_int*  nrows = new s_int [nlevel];
 s_int*  ncols = new s_int [nlevel];

 SetSizeLevel( nrow, ncol, nlevel, nrows, ncols );
 tree = new TSBN( nlevel, nrows, ncols, ncpt,
                  nparents, nchildren, nclass,
                  "code" );

} // end of: CodeTSBN::CodeTSBN

////////////////////////////////////////////////////////////
//
//
////////////////////////////////////////////////////////////
CodeTSBN::~CodeTSBN()
   {
   }




////////////////////////////////////////////////////////////
//
//  void  CodeTSBN::CodeTruncatedTree( char* fhe, 
//                                     float*** probs, 
//                                     char* fcodecost)
//  Description: Calculate the coding cost along the 
//  truncated tree.
// 
//  Xiaojuan Feng  12/01/99
//
////////////////////////////////////////////////////////////
void  CodeTSBN::CodeTruncatedTree( char* fhe, 
                                   float*** probs, 
                                   char* fcodecost)
{
 s_int      i, j, k;
 s_int      nrow = tree->num_rows[tree->num_levels-1];
 s_int      ncol = tree->num_cols[tree->num_levels-1];
 s_int      nclass = tree->num_classes;
 int	    nsize = nrow*ncol;
 s_int*     images1 = new s_int [nsize];
 s_int*     labels1 = new s_int [nsize]; 
 int*       num_missing_pixel = new int [num_image];
 float      temp;
 float**    codingcost = new float * [num_image];
 FILE       *fp1;
   
 
 // set the priors an d CPT to b.net
 //tree->SetCPT1( probs );
 tree->SetCPT6( probs );
 //tree->set_cond_prob_3( probs );

 if ((fp1=fopen( fhe, "r")) == NULL)
    errors("file_hard doesn't exist\n"); 
 
 
 for (i=0; i<num_image; ++i) {
     
     // if ( i!=13 ) continue;
     printf("image: %d\n", i);

     // input
     for (j=0; j<nsize; ++j) {
          fscanf(fp1, "%f ", &temp);
          images1[j] = temp;
     }

     ChangeFormat( images1, labels1 );
     num_missing_pixel[i] = GetNumMissingPixel( labels1);

     tree->CalculatePrior( 6 );
     tree->SetEvidence( labels1 );
     tree->CalculateLambdaMargProb();
     codingcost[i]=tree->OutputLogMargProb();
     
     for (j=0; j<tree->num_levels; ++j) 
         codingcost[i][j] = 
             -codingcost[i][j]*1.4427/
              (nrow*ncol-num_missing_pixel[i]);
 }
 
 fclose(fp1);


 ofstream out(fcodecost);
 for (i=0; i<num_image; ++i) {
     for (j=0; j<tree->num_levels; ++j) 
         out << codingcost[i][j] << " ";
     out << "\n";
 }
 out.close();
} // end of CodeTSBN::CodeTruncatedTree()

     

