///////////////////////////////////////////////////////////
// 
// Filename:    hipsCD7.cc
// Description 
// Implementation file for class Hips. This class is for 
// working on the BA image database (on CD).
//
// Copyright 1998 (c)    Xiaojuan Feng 
// 
///////////////////////////////////////////////////////////

#include  <stdlib.h>
#include  <string.h> 
#include  <math.h>
#include  <iostream.h>
#include  <fstream.h>

#include  "hipsCD7.hh"
#include  "util.hh"

///////////////////////////////////////////////////////////
// 
// Hips::Hips(char* FNAME)
// Description
// The constructor of class: Hips.
// FNAME -- The file name for an image been processing.
// 
///////////////////////////////////////////////////////////
Hips::Hips(char* fname)
{

ReadRGBImage(fname);
ReadLabelFile(fname);
ReadSegmentedImage(fname);

num_col = hips_header.ncol;
num_row = seg_hips_header.nrow-2;

// The following functions are used to display  images
// (colour images, label images, segmentation images by
// calling xv.)
//
//hips2ppm(fname, rgb);
//hips2pgm(fname);
//seg2ppm(fname);

} 


///////////////////////////////////////////////////////////
// 
// Hips::~Hips(char* FNAME)
// Description
// The disconstructor of class: Hips.
// 
///////////////////////////////////////////////////////////
Hips::~Hips()
{
}


///////////////////////////////////////////////////////////
// 
// Hips::GetNumRow();
//
///////////////////////////////////////////////////////////
int  Hips::GetNumRow()
{ 
 return num_row;
} // endo of Hips::GetNumRow()


///////////////////////////////////////////////////////////
// 
// Hips::GetNumRow();
//
///////////////////////////////////////////////////////////
int  Hips::GetNumCol()
{
  return num_col;
}// end of  Hips::GetNumCol()

///////////////////////////////////////////////////////////
//
// Hips::ReadRGBImage(char* fname)
// Aim: 	Read the original colour image
//
///////////////////////////////////////////////////////////
void  Hips::ReadRGBImage(char* fname)
{
char* 		fname1=new char[strlen(fname)+20];
FILE 		*fp;

 strcpy( fname1, fname);
 strcat( fname1, ".col.hips");
 fp = fopen(fname1, FILE_READ_BIN);
 if ( !fp ) 
    Error1("Error in openning file: *.col.hips !\n");

 ReadHeader(fp, &hips_header);
 AllocateRGB(&rgb);
 ReadPixelRGB(fp);
 
 fclose(fp);
 delete [] fname1;
}  // end of ReadRGBImage()


///////////////////////////////////////////////////////////
//
//  Hips::ReadHeader(FILE* fp)
//  Aim:	Read the header of a hips image
//
///////////////////////////////////////////////////////////
void	Hips::ReadHeader(FILE* fp, 
                         HipsHeader* hips_header)
{
char ptr[LINESIZE];

  // Read and toss orig_name.
  GetLine(fp, ptr, LINESIZE-1);
  // Read and toss seq_name.
  GetLine(fp, ptr, LINESIZE-1);
  // Read num_frame.
  GetLine(fp, ptr, LINESIZE-1);
  hips_header->num_frame = atoi(ptr);
  // Read and toss orig_date.
  GetLine(fp, ptr, LINESIZE-1);
  // Read rows.
  GetLine(fp, ptr, LINESIZE-1);
  hips_header->nrow = atoi(ptr );
  // Read cols.
  GetLine(fp, ptr, LINESIZE-1);
  hips_header->ncol = atoi(ptr );
  // Read bits_per_pixel.
  GetLine(fp, ptr, LINESIZE-1);
  hips_header->bits_per_pixel = atoi(ptr );
   // Read bit_packing.
  GetLine(fp, ptr, LINESIZE-1);
  hips_header->bit_packing = atoi(ptr );
  // Read pixel_format.
  GetLine(fp, ptr, LINESIZE-1);
  hips_header->pixel_format = atoi(ptr ); 
  // Now read and toss lines until we get 
  // one with just a period.
  do{
    GetLine(fp, ptr,LINESIZE-1);
  }while (strcmp(ptr, ".\n" ) != 0 );
  return;
} // end of ReadHeader() 


///////////////////////////////////////////////////////////
//
//   Hips::GetLine( FILE* fp, char* ptr, int size )
//   Aim:	Read a line 
//
///////////////////////////////////////////////////////////
void Hips::GetLine( FILE* fp, 
                    char* ptr, 
                    int size )
{
  if ( fgets(ptr,size,fp) == NULL )   
     Error1("Error in reading Hips header !\n");

  return;
} // end of GetLine()


///////////////////////////////////////////////////////////
//
// Hips::AllocateRGB()
// Aim:	Set the allocation for the pixels of a colour image
//
///////////////////////////////////////////////////////////
void Hips::AllocateRGB(Rgb* rgb)
{
int  i;
int  nrow=hips_header.nrow,  ncol=hips_header.ncol;

  rgb->red   = new un_char* [nrow];
  rgb->green = new un_char* [nrow];
  rgb->blue  = new un_char* [nrow];

  for(i=0; i< nrow;i++){ 
    rgb->red[i]   = new un_char[ncol];
    rgb->green[i] = new un_char[ncol];
    rgb->blue[i]  = new un_char[ncol];
  }
 
  return;
} // end of AllocateRGB()



///////////////////////////////////////////////////////////
//
//  Hips::ReadPixelRGB(FILE *fp)
//  Aim:	Read the pixels of a colour image
//
///////////////////////////////////////////////////////////
void Hips::ReadPixelRGB(FILE *fp)
{
int i,j, nrow=hips_header.nrow, ncol=hips_header.ncol;

if ( (hips_header.num_frame == 3) && 
     (hips_header.pixel_format == 0) ) {
   for (i=0; i<nrow; i++)
       FileReadBin((rgb.red[i]),
                   ncol*sizeof(un_char),1,fp,
                   "Error in read_pixel_rgb !\n" );
   for (i=0; i<nrow; i++)
       FileReadBin((void*)(rgb.green[i]),
                   ncol*sizeof(un_char),1,fp,
                   "Error in read_pixel_rgb !\n");
   for (i=0; i<nrow; i++)
       FileReadBin((void*)(rgb.blue[i]),
                   ncol*sizeof(un_char),1,fp,
                   "Error in read_pixel_rgb !\n");
}
else if ((hips_header.num_frame==1) && 
         (hips_header.pixel_format==35)) {
        for (i=0; i<nrow; i++)
        for (j=0; j<ncol; j++) {
            FileReadBin(&rgb.red[i][j],  
                        1*sizeof(un_char),1,fp,
                        "Error in read_pixel_rgb !\n");
            FileReadBin(&rgb.green[i][j],   
                        1*sizeof(un_char),1,fp,
                        "Error in read_pixel_rgb !\n");
            FileReadBin(&rgb.blue[i][j],   
                        1*sizeof(un_char),1,fp,
                        "Error in read_pixel_rgb !\n");   
        }
}                                          

return;
} // end of ReadPixelRGB()


///////////////////////////////////////////////////////////
//
//  Hips::read_label_image(char* fname)
//  Aim:	Read the label file
//
///////////////////////////////////////////////////////////
void   Hips::ReadLabelFile(char* fname)
{
 char* 		fname1=new char[strlen(fname)+20];
 FILE 		*fp;

 strcpy( fname1, fname);
 strcat( fname1, ".obj.def");
 fp = fopen(fname1, FILE_READ_BIN);
 if ( !fp ) 
 Error1("Error in openning file: *.obj.def !\n");
 
 // number of regions composing the image
 nregs = ReadLab3Header(fp);
 idl_regtbl = new LabelMape[nregs];

 // map of the labels
 ReadLabel(fp);
 
 fclose(fp);
 delete [] fname1;
} // end of ReadLabelFile()


///////////////////////////////////////////////////////////
//
//  Hips::ReadLab3Header(FILE *fp)
//  Aim:	Read the header of a label file
//
///////////////////////////////////////////////////////////
int  Hips::ReadLab3Header(FILE *fp)
{
 int 	status, validity, tmp_nregions;

 FileReadBin(&status, sizeof(int), 1, fp,
             "Label File Read Failed ! \n");
 if (status != LAB_MAGIC)
     Error1("Label File Read Failed !\n");
 FileReadBin(&validity, sizeof(int), 1, fp,
             "Label File - Bad Format !\n");

 //For backward compatibility check validity flag exists:
 if ( (validity==0)||(validity==1) ) {
    //read number of regions in this labelmap
    FileReadBin(&tmp_nregions, sizeof(int), 1, fp,
                "Label File - Bad Format !\n");
 }
 else{
    //For backward compatibility with v1.0 label files:
    tmp_nregions = validity;
    validity = -1;
 }
 if (tmp_nregions<=0)
 Error1("Label File - Bad Format\n");
 FileReadBin(&status, sizeof(int), 1, fp,
             "Label File - Bad Format\n");

 #ifdef DEBUG
 if (DEBUGLEVEL > 2) {
    fprintf(stderr, "Loading labelmap file...\n");
    fprintf(stderr, "Label type:");
 }
 #endif

 switch (status) {
        case LABELMAP_OBJECTS_IDENTITIES:
             #ifdef DEBUG
             if (DEBUGLEVEL > 2)
                fprintf(stderr, "Object identities\n");
             #endif
             break;

        default: Error1("Unknown\n");
                 break;
            
        #ifdef DEBUG
        if(DEBUGLEVEL > 2)  fprintf(stderr, "Unknown\n");
        #endif
 }
 
 return(tmp_nregions);
} // end of ReadLab3Header()


///////////////////////////////////////////////////////////
//
//  Hips::ReadLabel(FILE *fp)
//  Aim:	Read the label fro each pixel
//
///////////////////////////////////////////////////////////
void  Hips::ReadLabel(FILE *fp)
{
int 	i;
int 	status, serialidx, nitems;
double  labelid;

for (i=0; i<nregs; i++){
    FileReadBin(&idl_regtbl[i].labelid, 
                sizeof(double),1,fp,
	        "Label File - Bad Format \n");
    FileReadBin(&idl_regtbl[i].labelidlist,
                SIZE_OF_DOUBLE_POINTER,1,fp,
		"Label File - Bad Format \n");
    FileReadBin(&idl_regtbl[i].serialidx,sizeof(int),1,fp,
	        "Label File - Bad Format \n");
    FileReadBin(&idl_regtbl[i].serialidxlist,
                SIZE_OF_INT_POINTER,1,fp,
		"Label File - Bad Format \n");
    FileReadBin(&idl_regtbl[i].qualifier,
                sizeof(double),1,fp,
	      	"Label File - Bad Format \n");
    nitems = 0;
    if (idl_regtbl[i].labelidlist != (double *) NULL){
       //This regions is multiply labelled:
       do{ //Read label and its serial index:
          FileReadBin(&labelid,sizeof(double),1,fp,
                      "Label File - Bad Format \n");
          status = (int)(((*(un_char*)&labelid) & 
                          STOP_CODE)
                         !=STOP_CODE);
          //Status = 1 <=> left != right;
          //Status = 0 <=> left  = right;
          //cout<<"labelid: "<<status<<endl;
          if (status != 0){
 	     //not at the end of higher order labels
	     nitems++;
	     FileReadBin(&serialidx,sizeof(int), 1,fp,
	  	         "Label File - Bad Format \n");
   	     //Read label qualifier (no serial index):
      	     FileReadBin(&labelid,sizeof(double), 1,fp,
		         "Label File - Bad Format \n");
          }
       } while (status!=0);
       idl_regtbl[i].labelidlist = (double *)(NULL);
       idl_regtbl[i].serialidxlist = (int *)(NULL);
    }

    #ifdef DEBUG
    if (DEBUGLEVEL > 2)
       fprintf(stderr, 
               "region %5d has serialidx %3d\
               (%2d extra labels).\n",
               i,rlme->serialidx,nitems);
    #endif
  }

  #ifdef DEBUG
  if (DEBUGLEVEL > 0)
     fprintf(stderr, "%d Regions in label Map\n",
             tmp_nregions);
  #endif
  
  return ;
} // end of ReadLabel()



///////////////////////////////////////////////////////////
//
//  Hips::ReadSegmentedImage(char* fname)
//  Aim:	Read the segmented image
//
///////////////////////////////////////////////////////////
void    Hips::ReadSegmentedImage(char* fname)
{
 char* 		fname1=new char[strlen(fname)+20];
 FILE*		fp;

 strcpy( fname1, fname);
 strcat( fname1, ".obj");
 fp = fopen(fname1, FILE_READ_BIN);
 if ( !fp ) Error1("Error in opening file: *.obj !\n");
 
 //Reading and tossing the header of the *.seg3.hips file
 ReadHeader(fp, &seg_hips_header);
 //Region contains the region to which each pixel
 //belongs to
 region = simatrix(seg_hips_header.nrow-2, 
                   seg_hips_header.ncol-2);
 //Reading the regions in the *.seg3.hips file
 ReadRegion(fp);
 
 fclose(fp);
 delete [] fname1;
} // end of Hips::ReadSegmentedImage()
 

///////////////////////////////////////////////////////////
// 
//  Hips::ReadRegion(FILE* fp)
//  Aim:	Read region ??
//
///////////////////////////////////////////////////////////
void  Hips::ReadRegion(FILE* fp)
{
 int 	i, nrow=seg_hips_header.nrow-2; 
 int    ncol=seg_hips_header.ncol-2;
 s_int* pix;

 pix = new s_int[ncol+1];               
 //The first row is the border.
 FileReadBin(pix, (ncol+1)*sizeof(s_int), 1, fp,
             "Error in reading the first row of\
             ssegmented image!\n"); 
 delete [] pix ;

 // For each row of the image (the last is tossed)
 pix = new s_int[2];
 for (i=0; i<nrow; i++) {
      // The first two pixels belong to the border 
      // of the previous and the actual row
      FileReadBin(pix, 2*sizeof(s_int),1,fp, 
                  "Error in ReadRegion 1 !\n");
      FileReadBin(region[i], ncol*sizeof(s_int),1,fp,
                  "Error in ReadRegion 2 !\n"); 
  }
  delete [] pix ;

  return;
} // end of Hips::ReadRegion() 



///////////////////////////////////////////////////////////
//
// Hips::FileReadBin(void* ptr,int size,int num,FILE* fp,char* msg)
// Aim:		Read a size ??
//
///////////////////////////////////////////////////////////
void Hips::FileReadBin(void* ptr,
                       int size,
                       int num,
                       FILE* fp,
                       char* msg)
{
  if (fread(ptr,size,num,fp)== NULL) Error1(msg);
  return;
} // end of FileReadBin() 



///////////////////////////////////////////////////////////
//
//  Hips::Error1(char* msg)
//  Aim:	Print error message and exit 
//
///////////////////////////////////////////////////////////
void  Hips::Error1(char* msg)
{
  cout << msg;
  exit(1);
}

///////////////////////////////////////////////////////////
//
//  Hips::hips2ppm(char*)
//  Aim:  	Change hips image into the format of ppm
//
///////////////////////////////////////////////////////////
void  Hips::hips2ppm(char* fname, Rgb rgb)
{
 int		ncolours;
 char*		fname1=new char[strlen(fname)+20];
 FILE*		fp;

 strcpy( fname1, fname);
 strcat( fname1, ".ppm");
 fp = fopen(fname1, FILE_WRITE_BIN);
 if ( !fp ) Error1("Errro in writing file: *.ppm !\n");

 ncolours = (int)(pow(2, hips_header.bits_per_pixel)-1);
 Writeppm(fp, rgb, ncolours);
 fclose(fp);
 ViewImage(fname1);

 delete [] fname1;
 return ;
} // end of hipstoppm()


///////////////////////////////////////////////////////////
//
//   Hips::Writeppm(FILE* fp)
//   Aim:	Write an image into file *fp in the ppm format
//
///////////////////////////////////////////////////////////
void  Hips::Writeppm(FILE*  fp, 
                     Rgb    rgb, 
                     int    ncolours)
{
 int 	i, j;
 int 	nrow=hips_header.nrow, ncol=hips_header.ncol;

 // Write the header of the  *.ppm 
 ppmHeader(fp, ncolours);

 // Write the colour pixels 
 for (i=0; i<nrow; i++) 
 for (j=0; j<ncol; j++){
      if (fwrite(&rgb.red[i][j],   
                 sizeof(char), 1, fp)==NULL)
         Error1("Error Writing ppmfile !\n");
      if (fwrite(&rgb.green[i][j], 
                 sizeof(char), 1, fp)==NULL)
         Error1("Error Writing ppmfile !\n");
      if (fwrite(&rgb.blue[i][j],  
                 sizeof(char), 1, fp)==NULL)
         Error1("Error Writing ppmfile !\n");
  }
    
  return;
}  // end of Writeppm()


///////////////////////////////////////////////////////////
//
//  Hips::ppmHeader(FILE* fp, int ncolours)
//  Aim:	Write the header of a *.ppm file
//
///////////////////////////////////////////////////////////
void	Hips::ppmHeader(FILE* fp, 
                        int   ncolours)
{
 int	 nrow=hips_header.nrow, ncol=hips_header.ncol;
 char    s[LINESIZE];

  if (fputs("P6\n#\n#\n", fp) == EOF) 
     Error1("Error Writing the header of ppmfile !\n");
  sprintf(s, "%d %d \n", ncol, nrow);
  if (fputs(s, fp) == EOF) 
     Error1("Error Writing the header of ppmfile !\n");
  sprintf(s, "%d \n", ncolours);
  if (fputs(s, fp) == EOF) 
     Error1("Error Writing the header of ppmfile !\n");
} // end of ppmHeader


///////////////////////////////////////////////////////////
//
//  Hips::ViewImage(char*)
//  Aim:   Display  an image
//
///////////////////////////////////////////////////////////
void  Hips::ViewImage(char* fname)
{
 char   command[LINESIZE];

  sprintf(command, "xv %s &", fname);
  system(command);
  
  return;
} // end of ViewImage


///////////////////////////////////////////////////////////
//
// Hips::hips2pgm(char* fname)
// Aim:	 Visualisation of one region of the image belonging to 
//       a specific class
//
///////////////////////////////////////////////////////////
void  Hips::hips2pgm(char* fname)
{
 int 	  i, j, ncolours, nrow=seg_hips_header.nrow-2; 
 int      ncol=seg_hips_header.ncol-2;
 char*	  fname1=new char[strlen(fname)+20];
 s_int 	  label, test;
 FILE* 	  fp;
 
 //Insert the label you want to visualize
 cout<<"Insert the class you want to " <<
       "visualize (0 <= label <= 7): "<<endl;
 cout<<"(To visualize all of the 7 classes input 12) ";
 cin >> label;

 // 
 strcpy( fname1, fname);
 strcat( fname1, ".pgm11");

 if (label <= 7) { // visalize one class  
     if ( !(fp = fopen(fname1, FILE_WRITE_BIN)) )
     Error1("Error in openning file: *.pgm11 ! \n");

     ncolours = 1;
     pgmHeader(fp, ncolours);
     for (i=0; i<nrow; i++)
     for (j=0; j<ncol; j++) {
         //if(opt==2) test = 
         //idl_regtbl[region[i][j]].serialidx;
         test = Label7(idl_regtbl[region[i][j]].serialidx);
         if (test == label)  
	    // Write the back ground
            if (fwrite("0", sizeof(char), 1, fp) == NULL)
               Error1("Error Writing pgm_black !\n");
            else 
	    // Write the front ground
            if (fwrite("1", sizeof(char), 1, fp) == NULL)
               Error1("Error Writing pgm_white !\n");
     }
     fclose(fp);
     ViewImage(fname1);
  }
  else { // visalize all the 11 classes
     Rgb image7;
     AllocateRGB(&image7);
     for (i=0; i<nrow; ++i)
     for (j=0; j<ncol; ++j) {
         test = Label7(idl_regtbl[region[i][j]].serialidx);
         // This should be changed to colour7
         Colour11(&image7.red[i][j], &image7.green[i][j],
                 &image7.blue[i][j], test);
     }
     hips2ppm(fname1, image7);
     DeleteAllocatedRGB(&image7);
  }

  delete [] fname1;
  return;
} // end of hips2pgm()



///////////////////////////////////////////////////////////
//
//  Hips::pgmHeader(FILE*, int)
//  Aim:	Write the header of ??
//
///////////////////////////////////////////////////////////
void  Hips::pgmHeader(FILE* fp, int ncolours)
{
 int   nrow=hips_header.nrow, ncol=hips_header.ncol;
 char  str[LINESIZE];

 if (fputs("P5\n#\n#\n", fp) == EOF)
     Error1("Error Writing the header of pgmfile !\n");
 sprintf(str, "%d %d \n", ncol, nrow);
 if (fputs(str, fp) == EOF)
     Error1("Error Writing the header of pgmfile !\n");
 if (ncolours > 255) ncolours = 255;
 sprintf(str, "%d \n", ncolours);
 if (fputs(str, fp) == EOF)
     Error1("Error Writing the header of pgmfile !\n");

 return;
} // end of Hips::pgmHeader()




///////////////////////////////////////////////////////////
//
// Hips::Label7 (s_int serialidx)
// Aim:		
//    31/03/98
//
///////////////////////////////////////////////////////////
s_int   Hips::Label7(s_int serialidx)
{
 s_int res;

  switch(serialidx){
    case Atmospheric_Phenomena:
    case AP_Clear_Sky:
    case AP_Cloud:
    case AP_Mist_Smoke:
    case AP_Other:
      res = CLOUDS ;
      break;
    case Landscape:
    case L_Stone_Rock:
    case L_Soil:
    case L_Vegetation:
    case L_V_Bush:
    case L_V_Grass:
    case L_V_Hedge:
    case L_V_Tree:
    case L_V_Other:
    case L_Water:
    case L_Other:
    case RO_RB_Grass:
    case RO_RB_Leaves:
    case RO_RS_Leaves:
    case RO_RS_Grass:
      res = VEGETATION ;
      break;
    case RO_RS_Marking:
    case RO_RS_M_Centre:
    case RO_RS_M_Edge:
    case RO_RS_M_Other:
      res = ROAD_SURFACE_MARKING ;
      break;
    case RO_Road_Surface:
    case RO_RS_Concrete:
    case RO_RS_Drain:
    case RO_RS_Gravel:
    case RO_RS_Mud:
    case RO_RS_Stone_Rock:
    case RO_RS_Tarmac:
    case RO_RS_Water:
    case RO_RS_Other:
    case Road_Object:
    case RO_Road_Border:
    case RO_RB_Gravel:
    case RO_RB_Mud:
    case RO_RB_Stone_Rock:
    case RO_RB_Tarmac:
    case RO_RB_Concrete:
    case RO_RB_Drain:
    case RO_RB_Kerb_Stone:
    case RO_RB_Water:
    case RO_RB_Other:
      res = ROAD_SURFACE_BORDER ;
      break;
    case Building:
    case B_Chimney:
    case B_Door:
    case B_Roof:
    case B_Wall:
    case B_Window:
    case B_Other:
    case RO_Bridge:
    case BO_Gate:
    case BO_Wall:
      res = BUILDING ;
      break;
    case BO_Fence:
    case RO_Road_Sign:
    case Electrical_Object:
    case EO_Electricity_Pylon:
    case EO_Lamp_Post:
    case EO_Telegraph_Pole:
    case EO_Wire:
    case EO_Other:
    case RO_Traffic_Signal:
    case RO_Other:
      res = STREET_FURNITURE ;
      break;
    case Mobile_Object:
    case MO_Bike:
    case MO_B_Body:
    case MO_B_Wheel:
    case MO_B_Other:
    case MO_Car:
    case MO_C_Body:
    case MO_C_Wheel:
    case MO_C_Window:
    case MO_C_Other:
    case MO_Person:
    case MO_Truck:
    case MO_T_Body:
    case MO_T_Wheel:
    case MO_T_Window:
    case MO_T_Other:
    case MO_Van:
    case MO_V_Body:
    case MO_V_Wheel:
    case MO_V_Window:
    case MO_V_Other:
    case MO_Other:
      res = MOBILE_OBJECT_CAR ;
      break;
    case Undefined_Other:  
      res = UNDEFINED;
      break;
    default:
      res = UNDEFINED;
      break;
  }

return(res);
} // end of Label7()




//////////////////////////////////////////////////////////////
//
//   Hips::Colour11(u_char*,u_char*,u_char*,int )
//   Aim:	Match the colour by the label
//
//////////////////////////////////////////////////////////////
void Hips::Colour11(un_char* red, un_char* green,
                   un_char* blue,int label)
{

switch(label){
  case CLOUDS:
    *red = R_LIGHT_BLUE;
    *green = G_LIGHT_BLUE;
    *blue = B_LIGHT_BLUE;
    break;
  case VEGETATION:
    *red = R_GREEN;
    *green = G_GREEN;
    *blue = B_GREEN;
    break;
  case ROAD_SURFACE_MARKING:
    *red = R_BRIGHT_YELLOW;
    *green = G_BRIGHT_YELLOW;
    *blue = B_BRIGHT_YELLOW;
    break;
  case ROAD_SURFACE_BORDER:
    *red = R_GREY;
    *green = G_GREY;
    *blue = B_GREY;
    break;
  case BUILDING:
    *red = R_BRICK_RED;
    *green = G_BRICK_RED;
    *blue = B_BRICK_RED;
    break;
  case STREET_FURNITURE:
    *red = R_BRIGHT_RED;
    *green = G_BRIGHT_RED;
    *blue = B_BRIGHT_RED;
    break;
  case MOBILE_OBJECT_CAR:
    *red = R_TASTEFUL_PINK;
    *green = G_TASTEFUL_PINK;
    *blue = B_TASTEFUL_PINK;
    break;
  case UNDEFINED:
    *red = R_BLACK;
    *green = G_BLACK;
    *blue = B_BLACK;
    break;
  case REJECTED:
    *red = R_BLACK;
    *green = G_BLACK;
    *blue = B_BLACK;
    break;
  default:
    printf("%d\t",label);
    *red = R_BLACK;
    *green = G_BLACK;
    *blue = B_BLACK;
    break;
}

return;
} // end of Colour11




///////////////////////////////////////////////////////////
//
// Hips::segtoppm(char* fname);
// Aim:   Visualisation of one region of the image
//
///////////////////////////////////////////////////////////
void   Hips::seg2ppm(char* fname)
{
 int  	 reg, i, j;
 int     nrow=hips_header.nrow, ncol=hips_header.ncol;
 char* 	 fname1=new char[strlen(fname)+20];
 char    str=255;
 FILE*	 fp;

 strcpy(fname1, fname);
 strcat(fname1, ".obj.ppm");
 if (!(fp = fopen(fname1,FILE_WRITE_BIN)))
     Error1("Error in openning file: *.seg.ppm ! \n");

 //Number of region which you want to visualize
 do{
    cout << "Number of Region (1<= n <= " 
         << nregs<<"):" << endl;
    cin  >> reg;
 } while((reg > nregs)||(reg < 1));
 ppmHeader(fp, 1);

 for (i=0; i<nrow; i++) 
 for (j=0; j<ncol; j++)
      if (region[i][j]==reg) {
	 //  Write the colour values for this pixel 
         if (fwrite(&rgb.red[i][j],   
                    sizeof(char), 1, fp)==NULL)
            Error1("Error Writing region_ppmfile !\n");
         if (fwrite(&rgb.green[i][j], 
                    sizeof(char), 1, fp)==NULL)
            Error1("Error Writing region_ppmfile !\n");
         if (fwrite(&rgb.blue[i][j],  
                    sizeof(char), 1, fp)==NULL)
            Error1("Error Writing region_ppmfile !\n");
      }
      else {
	 // Write white colour for this pixel
         if (fwrite(&str, sizeof(char), 1, fp)==NULL) 
            Error1("Error Writing region_white_ppmfile");
         if (fwrite(&str, sizeof(char), 1, fp)==NULL) 
            Error1("Error Writing region_white_ppmfile");
         if (fwrite(&str, sizeof(char), 1, fp)==NULL) 
            Error1("Error Writing region_white_ppmfile");
      }

  fclose(fp);

  ViewImage(fname1);
  delete [] fname1;
  return;
} // end of Hips::seg2ppm()



///////////////////////////////////////////////////////////
//
//  Hips::DeleteAllocatedRGB(Rgb* rgb)
//  Aim: Delete the allocation for structure rgb
//
///////////////////////////////////////////////////////////
void Hips::DeleteAllocatedRGB(Rgb* rgb)        
{

 delete [] rgb->blue;
 delete [] rgb->red;
 delete [] rgb->green;

  return;
} // end of DeleteAllocatedRGB





///////////////////////////////////////////////////////////
// 
//  Hips::get_virtual_evidence(char* fname)
//  Aim:  See Node 08/06/97
//
///////////////////////////////////////////////////////////
void  Hips::GetVE (char* fname, int snrow)
{
 int      i, j, index;
 int	  nrow=num_row, ncol=num_col;
 char*    fname1=new char[strlen(fname)+20];
 int*     num=new int[MAX_LABELS_OUTPUT];
 double** means=dmatrix0(MAX_LABELS_OUTPUT, 3);
 double** vars =dmatrix0(MAX_LABELS_OUTPUT, 3);

 // Obtain the sample means and variances
 for (i=0; i<MAX_LABELS_OUTPUT; ++i) num[i] = 0;
 for (i=0; i<nrow; ++i)
 for (j=0; j<ncol; ++j) {
     index = Label7(idl_regtbl[region[i][j]].serialidx);
     ++num[index];
     means[index][0] += rgb.red[i][j];
     means[index][1] += rgb.green[i][j];
     means[index][2] += rgb.blue[i][j];
     vars[index][0]  += pow((double)rgb.red[i][j],   2.0);
     vars[index][1]  += pow((double)rgb.green[i][j], 2.0);
     vars[index][2]  += pow((double)rgb.blue[i][j],  2.0);
 }

 for (i=0; i<MAX_LABELS_OUTPUT; ++i) {
     if (num[i] !=0 ) {
        means[i][0] /= num[i];
        means[i][1] /= num[i];
        means[i][2] /= num[i];
        vars[i][0]  = vars[i][0]/num[i] - 
                      pow(means[i][0], 2.0);
        vars[i][1]  = vars[i][1]/num[i] - 
                      pow(means[i][1], 2.0); 
        vars[i][2]  = vars[i][2]/num[i] - 
                      pow(means[i][2], 2.0);
     }
 }
 
 // Obtain the virtual evidence: suppose the colour values
 // are Gaussian distributed
 strcpy(fname1, fname);
 strcat(fname1, ".ve");
 ofstream out(fname1);
 out << snrow << " " << snrow << endl;
 int    k;
 double prob;
 for (i=0; i<snrow; ++i) {
   for (j=0; j<snrow; ++j) {
     for (k=0; k<MAX_LABELS_OUTPUT; ++k) {
         if ((vars[k][0]<ZERO) ||
             (vars[k][1]<ZERO) ||
             (vars[k][2]<ZERO)) {
            prob = 0.0;
         }
         else {
            prob = 1/sqrt(2*PI*vars[k][0]*
                          vars[k][1]*vars[k][2])*
                   exp( -0.5*(
                   pow((rgb.red[i][j]-means[k][0]),2.0)/
                   vars[k][0]+
                   pow((rgb.green[i][j]-means[k][1]),2.0)/
                   vars[k][1]+
                   pow((rgb.blue[i][j]-means[k][2]),2.0)/
                   vars[k][2]));
         }
         out << prob << " ";
     }
     out <<"  ";
   }
   out << "\n";
 }
 out.close();      

 delete []  fname1;
 delete []  num;
 delete []  means;
 delete []  vars;
} // end of Hips::get_virtual_evidence();




///////////////////////////////////////////////////////////
//
//  void  Hips::ViewOutput(char* fname)
//
///////////////////////////////////////////////////////////
void  Hips::ViewOutput(char* fname)
{
 int	i, j, index, nrow, ncol;
 Rgb    image11;
 ifstream in(fname);

 AllocateRGB(&image11);
 in >> nrow >> ncol;
 for (i=0; i<nrow; ++i)
 for (j=0; j<ncol; ++j) {
     in >> index;
     Colour11(&image11.red[i][j], &image11.green[i][j],
              &image11.blue[i][j], index);
 }
 in.close();
 
 hips_header.nrow=nrow;
 hips_header.ncol=ncol;
 hips2ppm(fname, image11);
 DeleteAllocatedRGB(&image11);
} // end of Hips::view_CausalNetwork_outpu



///////////////////////////////////////////////////////////
//
//  Hips::DeleteAllocation()
//
///////////////////////////////////////////////////////////
void  Hips::DeleteAllocation()
{
 free(region);
 delete [] idl_regtbl;
 delete [] rgb.blue;
 delete [] rgb.green;
 delete [] rgb.red;
} // end of Hips::DeleteAllocation


//////////////////////////////////////////////////////////
//
//  Hips::GetNumClass()
//  Aim: To calculate the numbers of pixels in classes
//
//////////////////////////////////////////////////////////
void 	Hips::GetNumClass(FILE** fp)
{
 int	i, j, label;
 int	nrow = num_row, ncol = num_col;

 // Initialise the arrays
 for (i=0; i<=MAX_LABELS_OUTPUT; ++i)
     num_class[i] = 0;
 //  Calculate the number of pixels in each class
 for (i=0; i<nrow; ++i)
 for (j=0; j<ncol; ++j) {
     label = Label7(idl_regtbl[region[i][j]].serialidx);
     ++num_class[label]; 
 }
 // output 
 for (i=0; i<=MAX_LABELS_OUTPUT; ++i) {
     fprintf(*fp, "%7d ", num_class[i]);
     if (fmod(i+1, 5) == 0) fprintf(*fp, "\n");
 }
} // end of Hips::GetNumClasss()


//////////////////////////////////////////////////////////
//
//  Hips::OutputRandomPixel(int NUM, int SEED)
//  Aim: To output NUM(or less) pixels for each class.
//	 The seed for generating random number is SEED.
//
/////////////////////////////////////////////////////////
void Hips::OutputRandomPixel(int num, unsigned int seed,
                                FILE** fp)
{
 int	**rand_nums, rand_num, nitem, item;
 int    i, j, label;
 int	nrow = num_row, ncol = num_col;
 int    comp(const void*, const void *);
 void   push(int*, int*, int);

 // Allocate memory
 rand_nums = imatrix( (MAX_LABELS_OUTPUT+1), num );
 // Create NUM different random numbers in the range of
 // the number of pixels in each class
 for (j=1; j<=MAX_LABELS_OUTPUT; ++j) {
     srand( seed+j );
     nitem = 0;
     while ((nitem<num)&&(nitem<num_class[j])) {
           rand_num = rand();
           item = (int)(num_class[j]*(float)rand_num/
                        (float)RAND_MAX);
           push(rand_nums[j], &nitem, item);
     }
     // Sorting the random numbers
     qsort(rand_nums[j], nitem, sizeof(int), comp);
 }


 // output the selected pixels to file fp
 int 	**indexs;
 indexs = imatrix((MAX_LABELS_OUTPUT+1), 2);

 for (i=0; i<=MAX_LABELS_OUTPUT; ++i) {
     indexs[i][0] = 0;
     indexs[i][1] = 0;
 }
 for (i=0; i<nrow; ++i)
 for (j=0; j<ncol; ++j) {
     label = Label7(idl_regtbl[region[i][j]].serialidx);
     if (label > 0 ) {
        if (indexs[label][0]==
            rand_nums[label][indexs[label][1]] ) {
           // output this pixel
           fprintf(*fp, "%3d %3d %3d %2d\n", rgb.red[i][j],
                   rgb.green[i][j], rgb.blue[i][j], label);
           ++indexs[label][1];
        }
        ++indexs[label][0];
     }
 } 
     
 free(indexs);
 free(rand_nums);

} // end of  Hips::OutputRandomPixel()


//////////////////////////////////////////////////////////
// 
//  push(int*  stock, int* nitem, int item
//  Modification:
// 	11/09/97
//
/////////////////////////////////////////////////////////
void 	push(int* stock, int* nitem, int item)
{
 s_int  flag=1;
 int	   i;

 for (i=0; i<*nitem; ++i) {
     if (stock[i] == item) {
         flag = 0;
         break;
     }
 }
 if ( flag ) {
     stock[*nitem] = item;
     *nitem = *nitem + 1;
 }
} // end of push


//////////////////////////////////////////////////////////
//
// int  comp(const void *i, const void *j)
//
//////////////////////////////////////////////////////////
int 	 comp(const void *i, const void *j)
{
 return *(int *)i - *(int *)j;
}


///////////////////////////////////////////////////////////
//
//  un_char* Hips::GetAPixelValue(int i, int j)
//  Aim: To obtain the pixel values for a specified pixel
//  27/10/97
//
///////////////////////////////////////////////////////////
un_char * Hips::GetAPixelValue(int i, int j)
{
 un_char  rgbs[3];
 
 rgbs[0] = rgb.red[i][j];
 rgbs[1] = rgb.green[i][j];
 rgbs[2] = rgb.blue[i][j];
 
 return rgbs;
} // end of Hips::GetAPixelValue()



///////////////////////////////////////////////////////////
//
//  Hips::ReduceLabelImage(s_int ntimes, FILE** fp)
//  Aim: To output the reduced NTIMES label image
//       -1: for unlabeled block
//       0-6: for the classes
//
//  Xiaojuan Feng   31/03/98
//
///////////////////////////////////////////////////////////
void Hips::ReduceLabelImage( s_int  ntimes,
                             FILE** fp)
{
 s_int	    i, j, i1, j1, k, ki, kj;
 s_int      label, index, nclass1=MAX_LABELS_OUTPUT+1;   
 s_int      nrow = num_row/ntimes;
 s_int      ncol = num_col/ntimes;
 s_int*     nums = new s_int [nclass1]; 
                // Add 1 for undefined labels

 // Cut a square image
 //if (nrow < ncol) ncol = nrow;
 //else nrow = ncol;

 for (i=0; i<nrow; ++i) {
     i1 = i*ntimes;
     for (j=0; j<ncol; ++j) {
         j1 = j*ntimes;
         // for this block
         // Find the numbers of each class
         for (k=0; k<nclass1; ++k)
             nums[k] = 0;
         for (ki=0; ki<ntimes; ++ki)
	 for (kj=0; kj<ntimes; ++kj) {
             label = Label7(idl_regtbl
                     [region[i1+ki][j1+kj]].serialidx);
             if (label>MAX_LABELS_OUTPUT) {
                printf("Label wrong: row=%d  col=%d \
                       label=%d\n",
                       i1+ki, j1+kj, label);
                exit(1);
             }
             if (label != 0 ) nums[label] += 1;
         }
         // Get the majority class in this block
         index = 0;
         for (k=1; k<nclass1; ++k) 
             if (nums[index] < nums[k]) 
                index = k;
         // output the result
         fprintf(*fp, "%2d ", (index-1));
     } 
     fprintf(*fp, "\n");
 }
 delete nums;
} // end of  Hips::ReduceLabelImage()




///////////////////////////////////////////////////////////
//
//  Hips::CalculateFeature( s_int ntimes, FILE** fp)
//  Aim: To calculate features of each subimages (ntimes*ntimes).
//       The format of output file is:
//           class label, f1, f2, ..., fn (subimage [0,0])
//           class label, f1, f2, ..., fn (subimage [0,1])
//          ....
//           
//
//  Xiaojuan Feng   23/12/97
//		24/04/98
///////////////////////////////////////////////////////////
void Hips::CalculateFeature( s_int  ntimes, 
                             FILE** fp)
{
 s_int	    i, j, i1, j1, k, ki, kj;
 s_int      label, index, nclass1=MAX_LABELS_OUTPUT+1;   
 s_int      nrow = num_row/ntimes;
 s_int      ncol = num_col/ntimes;
 s_int*     nums = new s_int [nclass1]; 
                // Add 1 for undefined labels
 int        size = ntimes*ntimes;
 float      temp, temp1;
 float      f2[32];
 float***   subimage;

 void       calculate_I23(float***, s_int, float*);
 void       calculate_hue(float***, s_int, float*);
 void       calculate_gldv(float***, s_int, float*); 
 // allocation 
 // The first three depths are r, g, b; 
 // the the forth is gray value;
 subimage = fmatrix( ntimes, ntimes, 4);

 // Cut a square image
 //if (nrow < ncol) ncol = nrow;
 //else nrow = ncol;
 for (i=0; i<nrow; ++i) {
     i1 = i*ntimes;
     for (j=0; j<ncol; ++j) {
         j1 = j*ntimes;
	 
         // for this block:
         // find the numbers of each class,
         // put the pixel values into a matrix.
         for (k=0; k<nclass1; ++k)
             nums[k] = 0;
         for (ki=0; ki<ntimes; ++ki)
	 for (kj=0; kj<ntimes; ++kj) {
             subimage[ki][kj][0] = rgb.red[i1+ki][j1+kj];
             subimage[ki][kj][1] = rgb.green[i1+ki][j1+kj];
             subimage[ki][kj][2] = rgb.blue[i1+ki][j1+kj];
             label=Label7(
             idl_regtbl[region[i1+ki][j1+kj]].serialidx);
             if (label>MAX_LABELS_OUTPUT) {
                printf("Label wrong: row=%d  col=%d \
                       label=%d\n",
                       i1+ki, j1+kj, label);
                exit(1);
             }
             if (label != 0 ) nums[label] += 1;
         }
         // Get the majority class in this block
         index = 0;
         for (k=1; k<nclass1; ++k) 
             if (nums[index] < nums[k]) 
                index = k;
         // output the result
         // if ( index == 0 ) 
         // printf("undefine: row=%3d col=%3d\n",i,j);
         fprintf(*fp, "%2d ", index);

         // Feature 1: the average gray value 
	 //            ( gray intensity )
         // Feature 2: the variance of gray value
         temp = 0;
         temp1 = 0;
         for (ki=0; ki<ntimes; ++ki)
	 for (kj=0; kj<ntimes; ++kj) {
             subimage[ki][kj][3] = 0;
             for (k=0; k<3; ++k) 
             subimage[ki][kj][3] += subimage[ki][kj][k];
             subimage[ki][kj][3] /= 3.0;
             
             temp += subimage[ki][kj][3];
             temp1 += pow((double)subimage[ki][kj][3],2.0);
         }
         temp /= size;
         temp1 = temp1/size-pow((double)temp, 2.0);

         fprintf(*fp, "%6.2f %10.4f ", temp, temp1);
         // Feature 3,4: locations
         fprintf(*fp, "%3d %3d ", i1, j1);  
         // Feature 5,6: hue
         calculate_hue( subimage, ntimes, f2 );
         fprintf(*fp, "%10.7f %10.7f ", f2[0], f2[1]);
         // Feature 7..38: GLDV(Gray level difference 
	 // Vector)
         calculate_gldv( subimage, ntimes, f2);
         for (k=0; k<32; ++k)
             fprintf(*fp, "%10.4f ", f2[k]);   
	     
         // Feature 39 and 40: based on colour 
         calculate_I23( subimage, ntimes, f2 );
         fprintf(*fp, "%10.7f %10.7f ", f2[0], f2[1]);
         fprintf(*fp, "\n");

      } 
  }

 // have to remove these two when I resuce the nclass 
 // fro 11 to 7. Why?  
 //delete [] nums;  
 //delete [] subimage;
 
} // end of  Hips::CalculateFeature()


///////////////////////////////////////////////////////////
//
//  Hips::CalculateNormFeature( float** npare, s_int ntimes, FILE** fp)
//  Aim: To calculate features of each subimages (ntimes*ntimes).
//       The format of output file is:
//           class label, f1, f2, ..., fn (subimage [0,0])
//           class label, f1, f2, ..., fn (subimage [0,1])
//          ....
//           
//
//  Xiaojuan Feng   23/12/97
//
///////////////////////////////////////////////////////////
void Hips::CalculateNormFeature( float** npare, 
                                 s_int   ntimes, 
                                 FILE**  fp)
{
 s_int	    i, j, i1, j1, k, ki, kj;
 s_int      label, index, nclass1=MAX_LABELS_OUTPUT+1;   
 s_int      nrow = num_row/ntimes;
 s_int      ncol = num_col/ntimes;
 s_int*     nums = new s_int [nclass1]; 
                // Add 1 for undefined labels
 int        size = ntimes*ntimes;
 float      temp, temp1;
 float      f2[32];
 float***   subimage;

 void       calculate_I23(float***, s_int, float*);
 void       calculate_hue(float***, s_int, float*);
 void       calculate_gldv(float***, s_int, float*); 
 // allocation 
 // The first three depths are r, g, b; 
 // the the forth is gray value;
 subimage = fmatrix( ntimes, ntimes, 4);

 // Cut a square image
 //if (nrow < ncol) ncol = nrow;
 //else nrow = ncol;
 for (i=0; i<nrow; ++i) {
     i1 = i*ntimes;
     for (j=0; j<ncol; ++j) {
         j1 = j*ntimes;
         // for this block:
         // find the numbers of each class,
         // put the pixel values into a matrix.
         for (k=0; k<nclass1; ++k)
             nums[k] = 0;
         for (ki=0; ki<ntimes; ++ki)
	 for (kj=0; kj<ntimes; ++kj) {
             subimage[ki][kj][0]=rgb.red[i1+ki][j1+kj];
             subimage[ki][kj][1]=rgb.green[i1+ki][j1+kj];
             subimage[ki][kj][2]=rgb.blue[i1+ki][j1+kj];
             label=Label7(
             idl_regtbl[region[i1+ki][j1+kj]].serialidx);
             if (label>MAX_LABELS_OUTPUT) {
                printf("Label wrong: row=%d  col=%d \
                        label=%d\n",
                        i1+ki, j1+kj, label);
                exit(1);
             }
             if (label != 0 ) nums[label] += 1;
         }
         // Get the majority class in this block
         index = 0;
         for (k=1; k<nclass1; ++k) 
             if (nums[index] < nums[k]) 
                index = k;

         // output the result
         //if ( index == 0 ) 
	 //printf("undefine: row=%3d col=%3d\n",i,j);
         fprintf(*fp, "%2d ", index);

         // Feature 1: the average gray value 
	 //             ( gray intensity )
         // Feature 2: the variance of gray value
         temp = 0;
         temp1 = 0;
         for (ki=0; ki<ntimes; ++ki)
	 for (kj=0; kj<ntimes; ++kj) {
             subimage[ki][kj][3] = 0;
             for (k=0; k<3; ++k) 
             subimage[ki][kj][3] += subimage[ki][kj][k];
             subimage[ki][kj][3] /= 3.0;
             
             temp += subimage[ki][kj][3];
             temp1 += pow((double)subimage[ki][kj][3],2.0);
         }
         temp /= size;
         temp1 = temp1/size - pow((double)temp, 2.0);


         // normise
         temp  = (temp-npare[0][0])/npare[0][1];
         temp1 = (temp1-npare[1][0])/npare[1][1];

         fprintf(*fp, "%e %e ", temp, temp1);
         // Feature 3,4: locations
         temp  = ((float)i1-npare[2][0])/npare[2][1];
         temp1 = ((float)j1-npare[3][0])/npare[3][1];
         fprintf(*fp, "%e %e ", temp, temp1);  
         // Feature 5,6: hue
         calculate_hue( subimage, ntimes, f2 );
         f2[0] = (f2[0]-npare[4][0])/npare[4][1];
         f2[1] = (f2[1]-npare[5][0])/npare[5][1];
         fprintf(*fp, "%e %e ", f2[0], f2[1]);

         // Feature 7..38: GLDV
	 // (Gray level difference Vector)
         calculate_gldv( subimage, ntimes, f2);
         for (k=0; k<32; ++k)
             fprintf(*fp, "%e ", 
             (f2[k]-npare[6+k][0])/npare[6+k][1]);   

         // Feature 39 and 40: based on colour 
         calculate_I23( subimage, ntimes, f2 );
         for (k=0; k<2; ++k)
             fprintf(*fp, "%e ", 
             (f2[k]-npare[38+k][0])/npare[38+k][1]);
	      
         fprintf(*fp, "\n");                  
         
     } 
  }
  
 
 //delete nums;
 //delete subimage;

} // end of  Hips::CalculateNormFeature()




///////////////////////////////////////////////////////////
//
//  void  calculate_hue( float*** subimage, s_int ntimes, float* f2 )
//  Aim: To calculate the hue features
//
//  23/12/97
///////////////////////////////////////////////////////////

void    calculate_hue( float*** subimage, 
                       s_int    ntimes, 
                       float*   f2 )
{
 int         i,j, k;
 float       c[3], temp=ntimes*ntimes;
 
 // obtain the average value
 for (k=0; k<3; ++k) c[k] =0;
 for (i=0; i<ntimes; ++i)
 for (j=0; j<ntimes; ++j)
 for (k=0; k<3; ++k)
     c[k] += subimage[i][j][k];
 for (k=0; k<3; ++k) c[k] /= temp;

 // calculate the two features
 float  phi;
 temp = 2*(float)sqrt((double)((c[0]-c[1])*(c[0]-c[1])+
                               (c[0]-c[2])*(c[1]-c[2])));
 if (temp > ZERO) 
    phi=(float)acos((double)(((c[0]-c[1])+(c[0]-c[2]))/
                             temp));
 else 
    phi = (float)acos(0.0);
 f2[0] = (float)cos((double)phi);
 f2[1] = (float)sin((double)phi);
} // end of  calculate_hue()



///////////////////////////////////////////////////////////
//
//  void  calculate_I23( float*** subimage, s_int ntimes, float* f2 )
//  Aim: To calculate the colour features, ref: Ohta'
//
//         23/04/98
///////////////////////////////////////////////////////////
void    calculate_I23( float*** subimage, 
                       s_int    ntimes, 
                       float*   f2 )
{
 int         i,j, k;
 
 // obtain the I2 and I3 for each pixel and sum them up
 for (k=0; k<2; ++k) f2[k] =0;
 for (i=0; i<ntimes; ++i)
 for (j=0; j<ntimes; ++j) {
     f2[0] += (subimage[i][j][0]-subimage[i][j][2]);
     f2[1] += (2*subimage[i][j][1]-
               subimage[i][j][0]-
               subimage[i][j][2])/2.0;
 }

 f2[0] /= ntimes*ntimes;
 f2[1] /= ntimes*ntimes;
} // end of  calculate_hue()



///////////////////////////////////////////////////////////
//
//  void  calculate_gldv( float*** subimage, 
//                                    s_int ntimes, 
//                                    float* f2 )
// Aim: To obtain the GLDV (gray level difference vector)  and calculate 
//      some statistics
//      
// 23/12/97
//
///////////////////////////////////////////////////////////
void    calculate_gldv( float*** subimage, 
                        s_int    ntimes, 
                        float*   f2 )
{
 int	      i, j, u;
 float        p[4][256], tol;
 
 // initialise
 for (i=0; i<4; ++i)
 for (j=0; j<256; ++j) p[i][j]=0;
 
 // k=0, h=1: theta=0
 tol = ntimes*(ntimes-1);
 for (i=0; i<ntimes; ++i)
 for (j=0; j<(ntimes-1); ++j) {
     u = (int)fabs((double)(subimage[i][j+1][3]-
                            subimage[i][j][3]));
     ++p[0][u];
     
 }
 for (j=0; j<256; ++j) p[0][j] /=tol;
 
 // k=-1, h=1: theta=45
 tol = (ntimes-1)*(ntimes-1);
 for (i=1; i<ntimes; ++i)
 for (j=0; j<(ntimes-1); ++j) {
     u = (int)fabs((double)(subimage[i-1][j+1][3]-
                            subimage[i][j][3]));
     p[1][u]++;
 }
 for (j=0; j<256; ++j) p[1][j] /=tol;
 // k=1, h=0: theta=90
 tol = (ntimes-1)*ntimes;
 for (i=1; i<ntimes; ++i)
 for (j=0; j<ntimes; ++j) {
     u = (int)fabs((double)(subimage[i-1][j][3]-
                            subimage[i][j][3]));
     p[2][u]++;
 }
 for (j=0; j<256; ++j) p[2][j] /=tol;
 // k=-1, h=-1: theta=135
 tol = (ntimes-1)*(ntimes-1);
 for (i=1; i<ntimes; ++i)
 for (j=1; j<ntimes; ++j) {
     u = (int)fabs((double)(subimage[i-1][j-1][3]-
                            subimage[i][j][3]));
     p[3][u]++;
 }
 for (j=0; j<256; ++j) p[3][j] /=tol;
 // Calculate the statistics: ref: Welch 89
 // contrast
 
 for (i=0; i<4; ++i) {
     f2[i] =0;
     for (j=0; j<256; ++j) 
         f2[i] += j*j*p[i][j];
     
 }
 // entropy
 for (i=0; i<4; ++i) {
     f2[4+i] = 0;
     for (j=0; j<256; ++j) 
         if (p[i][j] > ZERO)
           f2[4+i] -= p[i][j]*log((double)p[i][j]);
 }
 // Local homogeneity
 for (i=0; i<4; ++i) {
     f2[8+i] = 0;
     for (j=0; j<256; ++j)
         f2[8+i] += p[i][j]/(1+j*j);
 }
 // Angular Second Moment
 for (i=0; i<4; ++i) {
     f2[12+i] = 0;
     for (j=0; j<256; ++j)
         f2[12+i] += p[i][j]*p[i][j];
 }
 // Mean
 for (i=0; i<4; ++i) {
     f2[16+i] = 0;
     for (j=0; j<256; ++j)
         f2[16+i] += j* p[i][j];
 }
 // Standard Deviation
 for (i=0; i<4; ++i) {
     f2[20+i] = 0;
     for (j=0; j<256; ++j)
         f2[20+i] += j*j* p[i][j];
     f2[20+i] = sqrt((double)f2[20+i]);
 }
 // Cluster  Shade
 for (i=0; i<4; ++i) {
     f2[24+i] = 0; 
     for (j=0; j<256; ++j)
         f2[24+i] += pow((double)(j-f2[16+i]),3.0)*p[i][j];
     if  (f2[20+i] < ZERO)
         f2[24+i] =0;
     else f2[24+i] = f2[24+i] /pow( (double)f2[20+i], 1.5);
 }
 // Cluster prominence
 for (i=0; i<4; ++i) {
     f2[28+i] = 0; 
     for (j=0; j<256; ++j)
         f2[28+i] += pow((double)(j-f2[16+i]),4.0)*p[i][j];
     if  (f2[20+i] < ZERO)
         f2[28+i] =-3;
     else f2[28+i] = f2[28+i] /pow( (double)f2[20+i], 2)-3;
 }

     
} // end of calculate_GLDV()



//////////////////////////////////////////////////////////
//
//  Hips::ReduceFuzzyImage(s_int ntimes, FILE** fp)
//  Aim: To output the reduced NTIMES label image
//
//  Xiaojuan Feng   22/01/98
//
/////////////////////////////////////////////////////////
void Hips::ReduceFuzzyImage( s_int ntimes, 
                             FILE** fp)
{
 s_int	i, j, i1, j1, k, ki, kj;
 s_int      label,  nclass1=MAX_LABELS_OUTPUT+1;   
 s_int      nrow = num_row/ntimes;
 s_int      ncol = num_col/ntimes;
 float*         nums = new float [nclass1]; 
 s_int      size = ntimes*ntimes;
                // Add 1 for undefined labels

 // Cut a square image
 //if (nrow < ncol) ncol = nrow;
 //else nrow = ncol;

 for (i=0; i<nrow; ++i) {
     i1 = i*ntimes;
     for (j=0; j<ncol; ++j) {
         j1 = j*ntimes;
         // for this block
         // Find the numbers of each class
         for (k=0; k<nclass1; ++k)
             nums[k] = 0;
         for (ki=0; ki<ntimes; ++ki)
	 for (kj=0; kj<ntimes; ++kj) {
             label = Label7(
             idl_regtbl[region[i1+ki][j1+kj]].serialidx);
             if (label>MAX_LABELS_OUTPUT) {
                printf("Label wrong: row=%d  col=%d\
                        label=%d\n",
                        i1+ki, j1+kj, label);
                exit(1);
             }
             nums[label] += 1;
         }
         // Get the frequence class in this block
         // if all pixels in the block are undefined 
         // set equal value to all classes.
         if (nums[0] == size)
            for (k=0; k<nclass1; ++k) 
                nums[k] = 1.0/7.0;
         else            
            for (k=0; k<nclass1; ++k) 
                nums[k] /= size;
         // output
         for (k=1; k<nclass1; ++k)
             fprintf(*fp, "%f ", nums[k]);
         fprintf(*fp, "\n");
     } 
 }
  

 delete nums;


} // end of  Hips::ReduceFuzzyImage()



///////////////////////////////////////////////////////////
//
//  Hips::CalculateSelectedNormFeature( 
//		float** npare, 
//		s_int* index,
//		s_int ntimes, 
//		FILE** fp)
//  Aim: To calculate features of each subimages (ntimes*ntimes).
//       The format of output file is:
//           class label, f1, f2, ..., fn (subimage [0,0])
//           class label, f1, f2, ..., fn (subimage [0,1])
//          ....
//           
//
//  Xiaojuan Feng   19/08/98
//
///////////////////////////////////////////////////////////
void Hips::CalculateSelectedNormFeature( float**  npare, 
					 s_int    ntimes,
					 FILE**   fp)
{
 s_int	    i, j, i1, j1, k, ki, kj;
 s_int      label, index, nclass1=MAX_LABELS_OUTPUT+1;   
 s_int      nrow = num_row/ntimes;
 s_int      ncol = num_col/ntimes;
 s_int*     nums = new s_int [nclass1]; 
                // Add 1 for undefined labels
 int        size = ntimes*ntimes;
 float      temp, temp1, f2[32];
 float***   subimage;

 void       calculate_I23(float***, s_int, float*);
 void       calculate_hue(float***, s_int, float*);
 void       calculate_gldv_selected(float***,s_int,float*);
 // allocation 
 // The first three depths are r, g, b; 
 // the  forth is gray value;
 subimage = fmatrix( ntimes, ntimes, 4);

 // Cut a square image
 //if (nrow < ncol) ncol = nrow;
 //else nrow = ncol;

 for (i=0; i<nrow; ++i) {
     i1 = i*ntimes;
     for (j=0; j<ncol; ++j) {
         j1 = j*ntimes;
         // for this block:
         // find the numbers of each class,
         // put the pixel values into a matrix.
         for (k=0; k<nclass1; ++k)
             nums[k] = 0;
         for (ki=0; ki<ntimes; ++ki)
	 for (kj=0; kj<ntimes; ++kj) {
             subimage[ki][kj][0] = rgb.red[i1+ki][j1+kj];
             subimage[ki][kj][1] = rgb.green[i1+ki][j1+kj];
             subimage[ki][kj][2] = rgb.blue[i1+ki][j1+kj];
             label = Label7(
             idl_regtbl[region[i1+ki][j1+kj]].serialidx);
             if (label>MAX_LABELS_OUTPUT) {
                printf("Label wrong: row=%d  col=%d\
                        label=%d\n",
                        i1+ki, j1+kj, label);
                exit(1);
             }
             if (label != 0 ) nums[label] += 1;
         }
         // Get the majority class in this block
         index = 0;
         for (k=1; k<nclass1; ++k) 
             if (nums[index] < nums[k]) 
                index = k;

         // output the result
         //if ( index == 0 ) 
	 //printf("undefine: row=%3d col=%3d\n",i,j);
         fprintf(*fp, "%2d ", index);

         // Feature 1: the average gray value 
	 //            ( gray intensity )
         // Feature 2: the variance of gray value
         temp = 0;
         temp1 = 0;
         for (ki=0; ki<ntimes; ++ki)
	 for (kj=0; kj<ntimes; ++kj) {
             subimage[ki][kj][3] = 0;
             for (k=0; k<3; ++k) 
             subimage[ki][kj][3] += subimage[ki][kj][k];
             subimage[ki][kj][3] /= 3.0;
             
             temp += subimage[ki][kj][3];
             temp1 += pow((double)subimage[ki][kj][3],2.0);
         }
         temp /= size;
         temp1 = temp1/size - pow((double)temp, 2.0);


         // normise
         temp  = (temp-npare[0][0])/npare[0][1];
         temp1 = (temp1-npare[1][0])/npare[1][1];

         fprintf(*fp, "%f %f ", temp, temp1);
         // Feature 3,4: locations
         temp  = ((float)i1-npare[2][0])/npare[2][1];
         //temp1 = ((float)j1-npare[3][0])/npare[3][1]; 
         // not selected feature
         fprintf(*fp, "%f  ", temp);  
         // Feature 5,6: hue
         calculate_hue( subimage, ntimes, f2 );
         f2[0] = (f2[0]-npare[4][0])/npare[4][1];
         f2[1] = (f2[1]-npare[5][0])/npare[5][1];
         fprintf(*fp, "%f %f ", f2[0], f2[1]);

         // Feature 7..38:
	 // GLDV(Gray level difference Vector)
         calculate_gldv_selected( subimage, ntimes, f2);
         for (k=0; k<14; ++k)
             fprintf(*fp, "%f ", 
             (f2[k]-npare[5+k][0])/npare[5+k][1]);   

         // Feature 39 and 40: based on colour 
         calculate_I23( subimage, ntimes, f2 );
         for (k=0; k<2; ++k)
             fprintf(*fp, "%f ", 
             (f2[k]-npare[19+k][0])/npare[19+k][1]);
	      
         fprintf(*fp, "\n");                           
     } 
  }
   
 //delete nums;
 //delete subimage;

} // end of  Hips::CalculateSelectedNormFeature()




///////////////////////////////////////////////////////////
//
//  void  calculate_gldv_selected( float*** subimage, 
//                                    s_int ntimes, 
//                                    float* f2 )
// Aim: To obtain the GLDV (gray level difference vector)  and calculate 
//      some statistics
//      
// 19/08/98
//
///////////////////////////////////////////////////////////
void    calculate_gldv_selected( float*** subimage, 
                                 s_int    ntimes, 
                                 float*   f2 )
{
 int	      i, j, u;
 float        p[4][256], tol;
 
 // initialise
 for (i=0; i<4; ++i)
 for (j=0; j<256; ++j) p[i][j]=0;
 
 // k=0, h=1: theta=0
 tol = ntimes*(ntimes-1);
 for (i=0; i<ntimes; ++i)
 for (j=0; j<(ntimes-1); ++j) {
     u = (int)fabs((double)(subimage[i][j+1][3]-
                            subimage[i][j][3]));
     ++p[0][u];
 }
 for (j=0; j<256; ++j) p[0][j] /=tol;
 
 // k=-1, h=1: theta=45
 tol = (ntimes-1)*(ntimes-1);
 for (i=1; i<ntimes; ++i)
 for (j=0; j<(ntimes-1); ++j) {
     u = (int)fabs((double)(subimage[i-1][j+1][3]-
                            subimage[i][j][3]));
     p[1][u]++;
 }
 for (j=0; j<256; ++j) p[1][j] /=tol;
 // k=1, h=0: theta=90
 tol = (ntimes-1)*ntimes;
 for (i=1; i<ntimes; ++i)
 for (j=0; j<ntimes; ++j) {
     u = (int)fabs((double)(subimage[i-1][j][3]-
                            subimage[i][j][3]));
     p[2][u]++;
 }
 for (j=0; j<256; ++j) p[2][j] /=tol;
 // k=-1, h=-1: theta=135
 tol = (ntimes-1)*(ntimes-1);
 for (i=1; i<ntimes; ++i)
 for (j=1; j<ntimes; ++j) {
     u = (int)fabs((double)(subimage[i-1][j-1][3]-
                            subimage[i][j][3]));
     p[3][u]++;
 }
 for (j=0; j<256; ++j) p[3][j] /=tol;
 // Calculate the statistics: ref: Welch 89
 // contrast 7-10
 for (i=0; i<4; ++i) {
     f2[i] =0;
     for (j=0; j<256; ++j) 
         f2[i] += j*j*p[i][j];
     
 }
 for (i=2; i<4; ++i) {
     f2[4+i] = 0;
     for (j=0; j<256; ++j)
         f2[4+i] += p[i][j]*p[i][j];
 }
 // Mean : 23, 24, 25, 26
 for (i=0; i<4; ++i) {
     f2[6+i] = 0;
     for (j=0; j<256; ++j)
         f2[6+i] += j* p[i][j];
 }
 // Standard Deviation: 27, 28, 29, 30
 for (i=0; i<4; ++i) {
     f2[11+i] = 0;
     for (j=0; j<256; ++j)
         f2[11+i] += j*j* p[i][j];
     f2[11+i] = sqrt((double)f2[11+i]);
 }     
} // end of calculate_GLDV_selected()

