/////////////////////////////////////////////////////////////
//
// Filename:	inputTSBN.cc
// Description:	Implementation file for the InputTSBN class.
//
// Copyright (c) 1998   Xiaojuan Feng
//
/////////////////////////////////////////////////////////////


#include <iostream.h>
#include <fstream.h>
#include <limits.h>
#include <stdlib.h>

#include "inputTSBN.hh"


////////////////////////////////////////////////////////////
//
// TrainTSBN::InputTSBN(s_int  nclass, 
//                      s_int  nlevel,
//                      s_int  nrow,   
//                      s_int  ncol,
//                      s_int  ncpt,   
//                      s_int  nimage)
// Description: The constructor of class: InputTSBN. 
// The parameters are:
// nclass -- Number of classes for each node.
// nlevel -- Number of levels in the TSBN.
// nrow   -- Number of rows at leaf level.
// ncol   -- Number of columns at the leaf level.
// ncpt   -- Number of CPTs in the TABN.
// nimage -- Number of images in using.
//
////////////////////////////////////////////////////////////
InputTSBN::InputTSBN(s_int  nclass, 
                     s_int  nlevel,
                     s_int  nrow,   
                     s_int  ncol,
                     s_int  ncpt,   
                     s_int  nimage)
{
 s_int*  nparents  = SetNumParent( nlevel );
 s_int*  nchildren = SetNumChildren( nrow, ncol, nlevel );
 s_int*  nrows = new s_int [nlevel];
 s_int*  ncols = new s_int [nlevel];
 num_image = nimage; 
 SetSizeLevel( nrow, ncol, nlevel, nrows, ncols );

 // Build up the TSBN
 tree = new TSBN( nlevel, nrows, ncols, ncpt,
                  nparents, nchildren, nclass );
  
} // end of: InputTSBN::TrainTSBN


////////////////////////////////////////////////////////////
//
// TrainTSBN::InputTSBN(s_int  nclass, 
//                      s_int  nlevel,
//                      s_int  nrow,   
//                      s_int  ncol,
//                      s_int  ncpt,   
//                      s_int  nimage)
// Description: The constructor of class: InputTSBN. 
// The parameters are:
// nclass -- Number of classes for each node.
// nlevel -- Number of levels in the TSBN.
// nrow   -- Number of rows at leaf level.
// ncol   -- Number of columns at the leaf level.
// ncpt   -- Number of CPTs in the TABN.
// nimage -- Number of images in using.
//
////////////////////////////////////////////////////////////
InputTSBN::InputTSBN(s_int  nclass, 
                     s_int  nlevel,
                     s_int  nrow,   
                     s_int  ncol,
                     s_int  ncpt,   
                     s_int  nimage,
                     s_int  try0)
{
 s_int*  nparents  = SetNumParent( nlevel );
 s_int*  nchildren = SetNumChildren( nrow, ncol, nlevel );
 s_int*  nrows = new s_int [nlevel];
 s_int*  ncols = new s_int [nlevel];
 num_image = nimage; 

 
 // Build up the TSBN

  
} // end of: InputTSBN::TrainTSBN




////////////////////////////////////////////////////////////
//
//  InputTSBN::~InputTSBN()
//  Description: The destructor of class: InputTSBN
//
////////////////////////////////////////////////////////////
InputTSBN::~InputTSBN()
{
}  




////////////////////////////////////////////////////////////
// 
// s_int* InputTSBN::ChangeFormat(s_int* images )
// Description:
// To reformat the 2D label image input intoa 1D
// node-index array. The following example shows the 
// relation between the two formats. In this example,
// the size of image is 4*6 pixels, each block is for 
// a pixel, the numbers in the blocks are the node-index
// in the 1D  evidence array.
//
// -------------------
// |0 | 1| 4| 5| 8| 9|
// -------------------
// | 2| 3| 6| 7|10|11|  Image
// -------------------
// |12|13|16|17|20|21|
// -------------------
// |14|15|18|19|22|23|
// -------------------
// 
////////////////////////////////////////////////////////////
s_int*  InputTSBN::ChangeFormat(s_int* images)
{
 int	i1hcol, n; 
 s_int	nlevel = tree->num_levels;
 s_int	nrow   = tree->num_rows[nlevel-1];
 s_int  ncol   = tree->num_cols[nlevel-1];
 s_int  nchild = tree->num_children[nlevel-2];
 s_int	i, j, i1, j1, im, jm, hcol, table[2][2]={0,1, 2,3};
 s_int*	labels = new s_int [nrow*ncol*num_image];

  // 
  if (nchild != 4)
     errors("The second last level should have 4 children. \
             (change_label)\n");
  int index;
  int index1;
  for (index=0; index<num_image; ++index) {
      index1 = index*nrow*ncol;
      hcol = ncol / 2;
      for (i=0; i<nrow; ++i) {
          i1 = i / 2;
          im = i - i1*2;
          i1hcol = i1*hcol;
          for (j=0; j<ncol; ++j) {
              j1 = j / 2;
              jm = j - j1*2;
              n = (i1hcol+j1)*4 + table[im][jm];
              labels[n+index1] = images[i*ncol+j+index1];
          }
      }
  } 
  return labels;
} // end of chagne_labels()


/////////////////////////////////////////////////////////////
// 
// InputTSBN::ChangeFormat(s_int* images,
//                          s_int* labels)
// Description:
// ?????????
/////////////////////////////////////////////////////////////
void InputTSBN::ChangeFormat(s_int* images,
                             s_int* labels)
{
 int   i1hcol, n; 
 s_int  nlevel = tree->num_levels;
 s_int	nrow   = tree->num_rows[nlevel-1];
 s_int  ncol   = tree->num_cols[nlevel-1];
 s_int  nchild = tree->num_children[nlevel-2];
 s_int	i, j, i1, j1, im, jm, hcol, table[2][2]={0,1, 2,3};
 
  // 
  if (nchild != 4)
     errors("The second last level should have 4 children.\
             (change_label)\n");

  int index;
  int index1;
  for (index=0; index<1; ++index) {
      index1 = index*nrow*ncol;
      hcol = ncol / 2;
      for (i=0; i<nrow; ++i) {
          i1 = i / 2;
          im = i - i1*2;
          i1hcol = i1*hcol;
          for (j=0; j<ncol; ++j) {
              j1 = j / 2;
              jm = j - j1*2;
              n = (i1hcol+j1)*4 + table[im][jm];
              labels[n+index1] = images[i*ncol+j+index1];
          }
      }
  }   
} // end of InputTSBN::chagne_labels()



////////////////////////////////////////////////////////////
// 
// float** InputTSBN::ChangeFormat( s_int** images )
// Description:
// To change the label image input into node-index array.
//
// ???????????      
//Chang the images into matrix for evidence format
// Modification:
//		19/01/98
//
////////////////////////////////////////////////////////////
float**  InputTSBN::ChangeFormat( float** images )
{
 int	 i1hcol, n; 
 s_int   nlevel = tree->num_levels;
 s_int	 nrow   = tree->num_rows[nlevel-1];
 s_int   ncol   = tree->num_cols[nlevel-1];
 s_int   nchild = tree->num_children[nlevel-2];
 s_int   nclass = tree->num_classes;
 s_int	 i,j,k,i1,j1,im,jm,hcol,table[2][2]={0,1, 2,3};
 float** labels;
  
  // allocation
  labels = fmatrix( nrow*ncol*num_image, nclass );
  // 
  if (nchild != 4)
     errors("The second last level should have 4 children.\
             (change_label)\n");

  int index;
  int index1;
  for (index=0; index<num_image; ++index) {
      index1 = index*nrow*ncol;
      hcol = ncol / 2;
      for (i=0; i<nrow; ++i) {
          i1 = i / 2;
          im = i - i1*2;
          i1hcol = i1*hcol;
          for (j=0; j<ncol; ++j) {
              j1 = j / 2;
              jm = j - j1*2;
              n = (i1hcol+j1)*4 + table[im][jm];
              for (k=0; k<nclass; ++k) 
                  labels[n+index1][k] = 
                  images[i*ncol+j+index1][k];
          }
      }
  } 

  return labels;
} // end of ChangeFormat()



////////////////////////////////////////////////////////////
// 
// float** InputTSBN::ChangeFormat(float** images, 
//                                    float** labels ) 
// Description:
// To change the label image input into node-index array.
//
// ???????????      
//Chang the images into matrix for evidence format
// Modification:
//		19/01/98
//
////////////////////////////////////////////////////////////
void InputTSBN::ChangeFormat( float** images, 
                              float** labels )
{
 int	i1hcol, n; 
 s_int  nlevel = tree->num_levels;
 s_int	nrow   = tree->num_rows[nlevel-1];
 s_int  ncol   = tree->num_cols[nlevel-1];
 s_int  nchild = tree->num_children[nlevel-2];
 s_int  nclass = tree->num_classes;
 s_int	i,j,k,i1,j1,im,jm,hcol,table[2][2]={0,1, 2,3};

 //
  if (nchild != 4)
     errors("The second last level should have 4 children.\
             (change_label)\n");

  int index;
  int index1;
  for (index=0; index<1; ++index) {
      index1 = index*nrow*ncol;
      hcol = ncol / 2;
      for (i=0; i<nrow; ++i) {
          i1 = i / 2;
          im = i - i1*2;
          i1hcol = i1*hcol;
          for (j=0; j<ncol; ++j) {
              j1 = j / 2;
              jm = j - j1*2;
              n = (i1hcol+j1)*4 + table[im][jm];
              for (k=0; k<nclass; ++k) 
                  labels[n+index1][k] = 
                  images[i*ncol+j+index1][k];
          }
      }
  }  
} // end of InputTSBN::ChangeFormat()


////////////////////////////////////////////////////////////
//
// s_int* InputTSBN::SetNumParent(s_int nlevel)
// Description:
// Set parameters for the structure of Belief network: 
// they can be fixed since we only  consider 
// quadtree-structured network at this moment, but 
// they may need to be changed as more complicated structures
// are used. 
//
////////////////////////////////////////////////////////////
s_int* InputTSBN::SetNumParent(s_int nlevel)
{
 s_int* nparents = new s_int [ nlevel ];
 s_int  i;
 nparents[0] = 0;
 for (i=1; i<nlevel; ++i) nparents[i] = 1;
 return nparents;
} // end of InputTSBN::SetNumParent()




////////////////////////////////////////////////////////////
//
// s_int* InputTSBN::SetNumChildren(s_int nrow, 
//                                  s_int ncol, 
//                                  s_int nlevel)
// Description:
// 
////////////////////////////////////////////////////////////
s_int* InputTSBN::SetNumChildren(s_int nrow, 
                                 s_int ncol, 
                                 s_int nlevel)
{
 s_int*	nchildren = new s_int [nlevel];
 if (( ncol==nrow)&&(nlevel>4))  
    nchildren[0] = 4;
 else if (( ncol != nrow)&&(nlevel>4)) 
    nchildren[0] = 6;
 else if (nlevel==3)
    nchildren[0] = 64*96;

 s_int i;
 for (i=1; i<nlevel; ++i) nchildren[i] = 4;
 nchildren[nlevel-1] = 0;

 return nchildren;
} // end of InputTSBN::SetNumChildren()





////////////////////////////////////////////////////////////
//
// void   InputTSBN::SetSizeLevel(s_int nrow, 
//                                 s_int ncol, 
//                                 s_int nlevel, 
//                                 s_int* nrows, 
//                                 s_int* ncols )
// Description:
// The two variables *nrows and *ncols in class TrainTSB
// are used to stored the information about the size of 
// image at each level in the TSBN. This subroutine is 
// used to assign the values to the two variables.
//
// Involved variables:
// *ncols,
// *ncols.
//
///////////////////////////////////////////////////////////
void 	InputTSBN::SetSizeLevel(s_int nrow, 
                                s_int ncol, 
                                s_int nlevel, 
                                s_int* nrows, 
                                s_int* ncols )
{
 // The numbers of nodes along row and column at each level
 s_int i;
 
 nrows[0] = 1;
 ncols[0] = 1;
 nrows[1] = 2;
 if ( ncol==nrow ) ncols[1] = 2;
 else    ncols[1] = 3;
 for (i=2; i<nlevel; ++i) {
     ncols[i] = ncols[i-1]*2;
     nrows[i] = nrows[i-1]*2;
 }
} // end of InputTSBN::SetSizeLevel()







////////////////////////////////////////////////////////////
// 
// int InputTSBN::GetNumMissingPixel(s_int* he )
// Description:
// To obtain the number of missing pixel in hard-evidences
// 
// 
////////////////////////////////////////////////////////////
int  InputTSBN::GetNumMissingPixel(s_int* he)
{
 s_int	nrow   = tree->num_rows[tree->num_levels-1];
 s_int  ncol   = tree->num_cols[tree->num_levels-1];
 int    nsize = nrow*ncol;
 int	i, num_missing_pixel=0;

 for (i=0; i<nsize; ++i)
   if (he[i] == -1) // undefined pixels labeled -1
      ++num_missing_pixel;

 return num_missing_pixel;
} // end of chagne_labels()







