/////////////////////////////////////////////////////////////
//
// Filename:	trainTSBN.cc
// Description:	Implementation file for the TrainTSBN class.
//
// Copyright (c) 1998   Xiaojuan Feng
//
/////////////////////////////////////////////////////////////


#include <iostream.h>
#include <fstream.h>
#include <limits.h>
#include <stdlib.h>

#include "trainTSBN.hh"


////////////////////////////////////////////////////////////
//
// TrainTSBN::TrainTSBN(s_int  nclass, 
//                      s_int  nlevel,
//                      s_int  nrow,   
//                      s_int  ncol,
//                      s_int  ncpt,   
//                      s_int  nimage)
// Description: The constructor of class: TrainTSBN. 
// The parameters are:
// nclass -- Number of classes for each node.
// nlevel -- Number of levels in the TSBN.
// nrow   -- Number of rows at leaf level.
// ncol   -- Number of columns at the leaf level.
// ncpt   -- Number of CPTs in the TABN.
// nimage -- Number of images in using.
//
////////////////////////////////////////////////////////////
TrainTSBN::TrainTSBN(s_int  nclass, 
                     s_int  nlevel,
                     s_int  nrow,   
                     s_int  ncol,
                     s_int  ncpt,   
                     s_int  nimage)
            : InputTSBN(nclass, 
                        nlevel,
                        nrow,   
                        ncol,
                        ncpt,   
                        nimage)
{
} // end of: TrainTSBN::TrainTSBN

////////////////////////////////////////////////////////////
//
//  TrainTSBN::~TrainTSBN()
//  Description: The destructor of class: TrainTSBN
//
////////////////////////////////////////////////////////////
TrainTSBN::~TrainTSBN()
{
}


////////////////////////////////////////////////////////////
//
//  void  TrainTSBN::TrainML( s_int*   IMAGES, 
//                            float*** PROBS, 
//                            float    TOL, 
//                            s_int    MAXITER)
//  Description: 
//  Training TSBN by maximising marginal likelihood of the 
//  hard evidence. EM algorithm is used for optimization. 
//  IMAGES  -- Array stored the labels.
//  PROBS   -- CPTS used in TSBN.
//  TOL     -- A measure of the precision required of 
//             the objective function at the solution. If 
//             the absolute difference between the objective 
//             function values between two successive iterations 
//             is less than TOL, then this condition 
//             is satisfied. 
//  MAXITER -- Maximum number of iterations.
//
//
//  See also
//  train.cc
//
////////////////////////////////////////////////////////////
void  TrainTSBN::TrainML( s_int*   images, 
                          float*** probs, 
                          float    tol, 
                          s_int    maxiter)
{
 char       fname[SIZE_FILENAME];
 s_int		i, j, k;
 s_int		nclass = tree->num_classes;
 s_int		ncpt = tree->num_cpts;
 s_int		iter =1;
 s_int*		labels = ChangeFormat( images );
 float		test = 100000;
 float*		sumlog = new float[maxiter];
 float**		logs = fmatrix(maxiter, num_image);
 FILE*		fp;

 delete images;
 do{
    printf("iter = %d\n", iter);

    Iteration( probs, labels, logs[iter]);

    sumlog[iter] = logs[iter][0];
    for (i=1; i<num_image; ++i) 
        sumlog[iter] += logs[iter][i];
    sumlog[iter] /= num_image;

    if (iter>0) test = fabs(sumlog[iter]-sumlog[iter-1]);
       
    fp = fopen("logs.dat", "w");
    for (i=0; i<(iter+1); ++i) {
        fprintf(fp, "%e \n", sumlog[i]);
	/*
        for (j=0; j<num_image; ++j)
            fprintf(fp, "%e ", logs[i][j]);
        fprintf(fp, "\n");
	*/
       }
    fclose(fp);
    
    if (iter >1) {
        sprintf(fname, "rm probs%d.dat", (iter-1));
        system(fname);
    }

    sprintf(fname, "probs%d.dat", iter);
    fp = fopen(fname, "w");
    for (j=0; j<nclass; ++j)
        fprintf(fp, "%e ", probs[0][0][j]);
    fprintf(fp, "\n");
    for (k=1; k<ncpt; ++k) 
    for (i=0; i<nclass; ++i) {
        for (j=0; j<nclass; ++j) 
            fprintf(fp, "%e ", probs[k][i][j]);
        fprintf(fp, "\n");   
    }       
    fclose(fp);

    ++iter;
 } while ((test>tol)&&(iter < maxiter));
} // end of TrainTSBN::TrainML()


////////////////////////////////////////////////////////////
//  void  TrainTSBN::TrainML( float**  IMAGES, 
//                            float*** PROBS, 
//                            float    TOL, 
//                            s_int    MAXITER )
//  Description: 
//  Training TSBN by maximising marginal likelihood of the 
//  virtual evidences. EM algorithm is used for optimization. 
//  IMAGES  -- Array stored the labels.
//  PROBS   -- CPTS used in TSBN.
//  TOL     -- A measure of the precision required of 
//             the objective function at the solution. If 
//             the absolute difference between the objective 
//             function values between two successive iterations 
//             is less than TOL, then this condition 
//             is satisfied. 
//  MAXITER -- Maximum number of iterations.
//
//
//  See also
//  train.cc
//
//
///////////////////////////////////////////////////////////
void  TrainTSBN::TrainML( float**  images, 
                          float*** probs, 
                          float    tol, 
                          s_int    maxiter)
{
 char       fname[SIZE_FILENAME];
 s_int 	    iter =0;
 s_int      i, j, k;
 s_int      nclass = tree->num_classes;
 s_int      ncpt = tree->num_cpts;
 float	    test = 100000;
 float      sumlog[200];
 float**    logs = fmatrix(200, num_image);
 float**    labels = ChangeFormat( images );
 FILE*      fp;

 delete images;

 do{
    printf("iter = %d\n", iter);

    Iteration( probs, labels, logs[iter]);

    sumlog[iter] = logs[iter][0];
    for (i=1; i<num_image; ++i) 
        sumlog[iter] += logs[iter][i];
    sumlog[iter] /= num_image;

    if (iter > 0)    
       test = fabs(sumlog[iter]-sumlog[iter-1]);
        
    fp = fopen("logs.dat", "w");
    for (i=0; i<(iter+1); ++i) {
        fprintf(fp, "%f ", sumlog[i]);
        for (j=0; j<num_image; ++j)
            fprintf(fp, "%f ", logs[i][j]);
        fprintf(fp, "\n");
       }
    fclose(fp);
    
    sprintf(fname, "probs%d.dat", iter);
    fp = fopen(fname, "w");
    for (j=0; j<nclass; ++j)
        fprintf(fp, "%f ", probs[0][0][j]);
    fprintf(fp, "\n");
    for (k=1; k<ncpt; ++k) 
    for (i=0; i<nclass; ++i) {
        for (j=0; j<nclass; ++j) 
            fprintf(fp, "%f ", probs[k][i][j]);
        fprintf(fp, "\n");   
    }       
    fclose(fp);

    ++iter;
    
 } while ((test>tol)&&(iter < maxiter));
} // end  of TrainTSBN::TrainML()





////////////////////////////////////////////////////////////
//
//  void TrainTSBN::Iteration( float***     probs,
//		               s_int*       labels,
//                             float*       logp)
//  Description:
//  
////////////////////////////////////////////////////////////
void  TrainTSBN::Iteration( float***     probs,
		            s_int*       labels,
                            float*       logp)
 {
  int 	    i, pi,pj,pk;
  s_int     nlevel = tree->num_levels;
  s_int     nrow = tree->num_rows[nlevel-1];
  s_int     ncol = tree->num_cols[nlevel-1];
  s_int     ncpt = tree->num_cpts;
  s_int     nclass = tree->num_classes;
  float***  newprobs = fmatrix(ncpt, nclass, nclass);
  float***  probs_out = fmatrix(ncpt, nclass, nclass);

  // set the priors an d CPT to b.net
  //tree->SetCPT1( probs );
  tree->SetCPT6( probs );
  //
  for (pk=0; pk<ncpt; ++pk)
  for (pi=0; pi<nclass; ++pi)
  for (pj=0; pj<nclass; ++pj) 
      newprobs[pk][pi][pj] = 0;
  
  // Iteration
  for (i=0; i<num_image; ++i) {
    //printf("image: %d\n", i);
      // set initial value for scale_d 
      tree->SetInitialValue();
      // set evidence
      tree->SetEvidence( (labels+i*nrow*ncol) );
      RunQuadtree( probs_out, &logp[i] );
      //
      for (pk=0; pk<ncpt; ++pk)
      for (pi=0; pi<nclass; ++pi)
      for (pj=0; pj<nclass; ++pj)
          newprobs[pk][pi][pj] += probs_out[pk][pi][pj];
  }

  // update the CPTs
  float temp;
  for (pk=0; pk<ncpt; ++pk)
  for (pi=0; pi<nclass; ++pi) {
      temp = newprobs[pk][pi][0];
      for (pj=1; pj<nclass; ++pj)
          temp += newprobs[pk][pi][pj];
      if (temp > ZERO ) 
         for (pj=0; pj<nclass; ++pj) 
             probs[pk][pi][pj] = newprobs[pk][pi][pj]/temp;
      else
         for (pj=0; pj<nclass; ++pj) 
             probs[pk][pi][pj] = 0;
  }
   
  delete newprobs;
  delete probs_out;

} // end of TrainTSBN::Iteration()
      




////////////////////////////////////////////////////////////
//
//    TrainTSBN::Iteration( float***	probs,
//		            s_int**     labels,
//                          float*      logp)
// 
////////////////////////////////////////////////////////////
void  TrainTSBN::Iteration( float*** probs,
		            float**  labels,
                            float*   logp)
{
  int 	    i, pi,pj,pk;
  s_int     nlevel = tree->num_levels;
  s_int     nrow = tree->num_rows[nlevel-1];
  s_int     ncol = tree->num_cols[nlevel-1];
  s_int     ncpt = tree->num_cpts;
  s_int     nclass = tree->num_classes;
  float***  newprobs = fmatrix(ncpt, nclass, nclass);
  float***  probs_out = fmatrix(ncpt, nclass, nclass);

  // set the priors an d CPT to b.net
  //tree->SetCPT11( probs );
  tree->SetCPT6( probs );

  //
  for (pk=0; pk<ncpt; ++pk)
  for (pi=0; pi<nclass; ++pi)
  for (pj=0; pj<nclass; ++pj) 
      newprobs[pk][pi][pj] = 0;
  
  // Iteration
  for (i=0; i<num_image; ++i) {
    //printf("image: %d\n", i);
      // set initial value for scale_d 
      tree->SetInitialValue();
      // set evidence
      tree->SetVirtualEvidence((labels+i*nrow*ncol));
      RunQuadtree( probs_out, &logp[i] );
  
      //
      for (pk=0; pk<ncpt; ++pk)
      for (pi=0; pi<nclass; ++pi)
      for (pj=0; pj<nclass; ++pj)
          newprobs[pk][pi][pj] += probs_out[pk][pi][pj];
  }

  // update the CPTs
  float temp;
  for (pk=0; pk<ncpt; ++pk)
  for (pi=0; pi<nclass; ++pi) {
      temp = newprobs[pk][pi][0];
      for (pj=1; pj<nclass; ++pj) 
          temp += newprobs[pk][pi][pj];
      if (temp > ZERO ) 
         for (pj=0; pj<nclass; ++pj) 
             probs[pk][pi][pj] = newprobs[pk][pi][pj]/temp;
      else
         for (pj=0; pj<nclass; ++pj) 
             probs[pk][pi][pj] = 0;
  }
   
  delete newprobs;
  delete probs_out; 
} // end of TrainTSBN::Iteration
      


////////////////////////////////////////////////////////////
//
// void TrainTSBN::RunQuadtree( float*** PROBS_OUT, 
//                                float*   LOGP)
// Description: 
// To pass message through a TSBN using the old CPTs,
// calculate the log likelihood log(P(evidence)), and
// calculate the estimated new CPTs.
// PROBS_OUT -- Estimated new CPTs,
// LOGP      -- Loglikelihood of the evidence.
//
////////////////////////////////////////////////////////////
void TrainTSBN::RunQuadtree( float*** probs_out, 
                             float*   logp)
{
 // Message passing
 tree->PassMessage();
 // estiamte the loglikelihood
 *logp = tree->Loglikelihood();
 // estimate the medium paramters
 tree->EstimateCPT( probs_out );
 
 return;
} // end of run_quadtree() 






///////////////////////////////////////////////////////////
//
//  void  TrainTSBN::LogMargLike( char*    fimage1, 
//                                float*** probs)
//  Description:
//  
///////////////////////////////////////////////////////////
/*
void  TrainTSBN::LogMargLike( char*    fimage1, 
                              float*** probs)
{
 s_int      i, j, k;
 s_int      nrow = tree->num_rows[tree->num_levels-1];
 s_int      ncol = tree->num_cols[tree->num_levels-1];
 s_int      nclass = tree->num_classes;
 int	    nsize = nrow*ncol;
 s_int*     images1 = new s_int [nsize];
 s_int*     labels1 = new s_int [nsize]; 
 float      logp[100];
 float      temp; 
 float**    conf = fmatrix(12,12);
 float          prior[11]={5.9417706e+01,   
                           25.8513751e+00,   
                           2.8735632e+01,   
                           15.8513751e+00,   
                           5.5034014e+00,   
                           7.5328885e+00,   
                           4.6711799e+00,   
                           4.2735043e+01,   
                           2.4330900e+01,   
                           3.2894737e+01,   
                           2.9154519e+01 }; 
 FILE        *fp1;
    
 // set the priors an d CPT to b.net
 //tree->SetCPT11( probs );
 tree->SetCPT6( probs );
 //tree->set_cond_prob_3( probs );

 if ((fp1=fopen( fimage1, "r")) == NULL)
    errors("file_hard doesn't exist\n"); 
 for (i=0; i<num_image; ++i) {
      printf("image: %d\n", i);
     
     // input
     for (j=0; j<nsize; ++j) {
          fscanf(fp1, "%f ", &temp);
          images1[j] = temp;
     }

     ChangeFormat( images1, labels1 );
 
     tree->SetInitialValue();
     tree->SetEvidence( labels1 );
     tree->PassMessage(); 
     logp[i] = tree->Loglikelihood();
 }
 
 fclose(fp1);

 fp1 = fopen("mlike_Bnet.dat","w");
 for (i=0; i<num_image; ++i)
     fprintf(fp1, "%f \n", logp[i]);
 fclose(fp1);
 
 delete [] images1;
 delete [] labels1;

} // end of TrainTSBN::LogMargLike()   

*/     




//////////////////////////////////////////////////////////
//
//  Function: CML training using Scaled Conjugate gradient.
//  void  TrainTSBN::TrainCMLscg( char*    fimage1, 
//			          char*    fimage2, 
//				  float*** probs, 
//                                double*  options)
//  Description 
//  Using a scaled conjugate gradients algorithm to find a 
//  local maximum of the conditional log likelihood 
//  function of the TSBN given hard_evidence 
//  ( stored in file: fimage1) and virtual evidence 
//  (stored in file: fimage2). The initial CPTs are given 
//  in 3D array probs.  The estimated CPTs are returned as 
//  probs. The optional parameters have the following
//  interpretations.
//
//  OPTIONS[0] is set to one to display error values; 
//
//  OPTIONS[1] is a measure of the absolute precision 
//             required for the value of PROBS at the 
//             solution. If the absolute difference 
//             between the values of the PROBS between 
//             two successive steps is less than OPTIONS[1],
//             then the condition is satisfied.
//
//  OPTIONS[2] is a measure of the precision required of 
//             the objective function at the solution. If 
//             the absolute difference between the objective 
//             function values between two successive steps 
//             is less than OPTIONS[2], then this condition 
//             is satisfied. Both this and the previous 
//             condition must be satisfied for termination.
//
//  OPTIONS[7] returns the conditional log likilihood 
//             values at the optimum values.
//
//  OPTIONS[9] returns the total number of function 
//             evaluations 
//             (including those in any line searchs).
//
//  OPTIONS[10] returns the total number of gradient 
//              evaluations.
//
//  OPTIONS[13] is the maximum number of Iterations.
//
//  OPTIONS[14] is set to 1 to check the gradients
//
//
//  See also
//  train.cc
//
///////////////////////////////////////////////////////////
void   TrainTSBN::TrainCMLscg( char*    fimage1, 
			       char*    fimage2, 
			       float*** probs, 
                               float*   options)
{
 // gradient check
 if (options[14] == 1)  
    CheckGradient(fimage1, fimage2, probs);

 int	nclass = tree->num_classes;
 int	ncpt = tree->num_cpts;
 char	fname[SIZE_FILENAME];
 s_int	i, j, k;
 float**	logs = fmatrix(options[13], (num_image+1));
 float	test;
 FILE*	fp;

 //  Initial some paramters
 s_int	    success =1;
 int        nsuccess = 1;// nsuccess counts number
                         // of succes.
 int        niter = 1;   // niter counts number of 
                         // Iterations.
 double     sigma0 = 1.0e-4; 
 double     lambda = 1.0;// Initial scale parameter.
 double     lambdamin = 1.0e-15; 
 double     lambdamax = 1.0e100;  
 float***   probs_val = fmatrix( ncpt, nclass, nclass );
 //  Initial function and gradient values
 double*    fold = new double[num_image+1];
 double*    fnew = new double[num_image+1];
 double*    f_temp = new double[num_image+1]; 
 float***   grad = fmatrix(ncpt,nclass, nclass);
 float***   gplus = fmatrix(ncpt,nclass, nclass);
 float***   gold  = fmatrix(ncpt,nclass, nclass);
 float***   gtemp = fmatrix(ncpt,nclass, nclass);
 void       updateProbs(int, int, int, float, 
                        float***, float***, float***);

 CalFunctionGradient( fimage1, fimage2, probs, fold, grad );
 ++options[10];
 for (i=0; i<=num_image; ++i)
      logs[0][i] = fold[i];
 // Initial search direction
 float***      srch = fmatrix(ncpt,nclass, nclass);
 setArray3D(ncpt, nclass, nclass, (float)-1.0, grad, srch);
 setArray3D(ncpt, nclass, nclass, (float)1.0, 
            probs, probs_val);
 // Main optimization loop
 double		mu;
 double		sigma;
 double		kappa;
 double         alpha;
 double         gamma;
 double         delta;
 double 	rho;
 double         beta;
 double         temp;
 do{
    // Calculate first and second directional derivatives
    if ( success == 1) {
       mu = productArray3D( ncpt, nclass, nclass, 
                            srch, grad);
       if (mu >= 0.0) {
          setArray3D( ncpt, nclass, nclass, 
                      (float)-1.0, grad, srch);
          mu = productArray3D( ncpt, nclass, nclass, 
                               srch, grad);
       }
       kappa = productArray3D( ncpt, nclass, nclass, 
                               srch, srch);
       sigma = sigma0 / sqrt(kappa);
       updateProbs(ncpt, nclass, nclass, (float)sigma, 
                   probs_val, srch, probs);
       CalFunctionGradient( fimage1, fimage2, probs, 
                            f_temp, gplus );  
       ++options[10];  
       // gamma = (srch * (gplus'-grad'))/sigma
       gamma = (double)productArray3D(ncpt, nclass, 
                                      nclass, srch, gplus);
       gamma -= (double)productArray3D(ncpt, nclass, 
                                       nclass, srch, grad);
       gamma /= sigma;
    } // if success == 1

    // Increase effective curvature and evaluate 
    // step size alpha
    delta = gamma + lambda * kappa;
    if (delta <= 0.0) {
       delta = lambda*kappa;
       lambda -= 2*gamma/kappa;
    }
    alpha = -mu/delta;
    
    // Calculate the comparison ratio
    updateProbs(ncpt, nclass, nclass, (float)alpha,
                probs_val, srch, probs);
    CalFunctionGradient( fimage1, fimage2, probs, 
                         fnew, gtemp );
    ++options[9];
    rho = 2 * (fnew[0]-fold[0]) / (alpha*mu);
    if (rho >= 0.0) success = 1;
    else success = 0;

    // Update the parameters to new location
    if (success == 1) {
       updateProbs(ncpt, nclass, nclass, (float)alpha, 
                   probs_val, srch, probs_val);
       ++nsuccess; 
    }
    setArray3D(ncpt, nclass, nclass, (float)1.0, 
               probs_val, probs);
    
    // display
    if (options[0] == 1) 
       printf("Iter: %4d  Error: %e  Scale: %e\n", 
              niter, fold[0], lambda);
    //
    if (success == 1) {
       // Test for termination 
       f_temp[0] = 0.0;
       for (k=0; k<ncpt; ++k)
       for (i=0; i<nclass; ++i)
       for (j=0; j<nclass; ++j)
           if ( f_temp[0] < fabs(srch[k][i][j]) )
              f_temp[0] = fabs(srch[k][i][j]);    
       if ((alpha*f_temp[0]<options[1]) && 
           (fabs(fnew[0]-fold[0])<options[2])) {
          options[7] = fnew[0];
          printf("Conditions are satistified.\n");
          return;
       }
       else {
	  // Update variables for new position
	  //fold = fnew;
          for (i=0; i<=num_image; ++i)  fold[i] = fnew[i];
          setArray3D(ncpt, nclass, nclass, 
                     (float)1.0, grad, gold);
          setArray3D(ncpt, nclass, nclass, 
                     (float)1.0, gtemp, grad);
          ++options[10];
          // If the gradient is zero then we are done.
          temp = (double)productArray3D(ncpt, nclass, 
                                        nclass, grad, grad);
          if (temp <= ZERO ){
             options[7] = fnew[0];
             printf("The gradient is zero.\n");
             return;
          }
       }
    } // end if success === 1

    // Adjust lambda according to comparison ratio
    if (rho < 0.25) {
       lambda = 4.0 * lambda;
       if (lambda > lambdamax) lambda = lambdamax;
    }
    else if (rho > 0.75) {
       lambda = 0.5*lambda;
       if (lambda < lambdamin) lambda = lambdamin;
    }

    // Re-compute search direction using Hestenes-Steifel 
    // formula, or re-start in direction of negative 
    // gradient after nparams steps.
    if (nsuccess == (ncpt*nclass*nclass)) {
       setArray3D(ncpt, nclass, nclass, 
                  (float)-1.0, grad, srch);
       nsuccess = 0;
    }
    else if (success == 1) { 
            beta  = productArray3D(ncpt, nclass, 
                                   nclass, gold, grad);
            beta -= productArray3D(ncpt, nclass, 
                                   nclass, grad, grad);
            beta /= mu;
            sumArray3D(ncpt, nclass, nclass, 
                       (float)-1.0, (float)beta, 
                       grad, srch, srch);
    }
    
    // Output the probs and fval
    for (i=0; i<=num_image; ++i)
        logs[niter][i] = fold[i];
    
    fp = fopen("logs.dat", "w");
    for (i=0; i<=niter; ++i) {
        fprintf(fp, "%e \n", -logs[i][0]);
	/*
        for (j=1; j<=num_image; ++j)
            fprintf(fp, "%e ", -logs[i][j]);
        fprintf(fp, "\n");
	*/
    }
    fclose(fp);
    
    if (niter >1) {
        sprintf(fname, "rm probs%d.dat", (niter-1));
       system(fname);
    }
    sprintf(fname, "probs%d.dat", niter);
    fp = fopen(fname, "w");
    for (j=0; j<nclass; ++j)
        fprintf(fp, "%e ", probs[0][0][j]);
    fprintf(fp, "\n");
    for (k=1; k<ncpt; ++k) 
    for (i=0; i<nclass; ++i) {
        for (j=0; j<nclass; ++j) 
            fprintf(fp, "%e ", probs[k][i][j]);
        fprintf(fp, "\n");   
    }       
    fclose(fp);

   if (niter > 1) test = fabs(-logs[niter][0]+logs[niter-1][0]);
    ++niter;   
         
 } while (niter < options[13]);
  
 // If we get here, then we havn't terminated in the 
 // given number of Iterations.
 options[7] = fold[0];
 if (options[0] >= 0)
    printf("Warning: Maximum number of Iterations has\
            been exceeded.\n");
 
} // end of TrainTSBN::TrainCMLscg()



//////////////////////////////////////////////////////////
//
// void TrainTSBN::CalFunctionGradient(char*    fimage1,
//				       char*    fimage2, 
//				       float*** probs, 
//				       double*  fval,
//				       float*** grad )	
// Description:
//
//////////////////////////////////////////////////////////
void TrainTSBN::CalFunctionGradient(char*    fimage1,
				    char*    fimage2, 
				    float*** probs, 
				    double*  fval,
				    float*** grad )
{
  int 	    i, j,k,pi,pj,pk;
  s_int     nlevel = tree->num_levels;
  s_int     nrow = tree->num_rows[nlevel-1];
  s_int     ncol = tree->num_cols[nlevel-1];
  s_int     ncpt = tree->num_cpts;
  s_int     nclass = tree->num_classes;
  int	    nsize = nrow*ncol;
  float     temp;
  s_int*    images1 = new s_int [nsize];
  s_int*    labels1 = new s_int [nsize]; 
  float**   images2 = fmatrix(nsize, (int)nclass);
  float**   labels2 = fmatrix(nsize, nclass);
  FILE      *fp1, *fp2;

  // set the priors and CPT to b.net
  //tree->SetCPT1( probs ); // One CPT on each level
  tree->SetCPT6( probs ); // six CPT on level 1
  // 
  float***      probs_out1 = fmatrix(ncpt, nclass, nclass);
                                     // m^n
  float***      probs_out2 = fmatrix(ncpt, nclass, nclass);
                                     // n^n
  double***     newprobs = dmatrix(ncpt, nclass, nclass);  
                                     // sum(m^n-n^n)
  for (pk=0; pk<ncpt; ++pk)
  for (pi=0; pi<nclass; ++pi)
  for (pj=0; pj<nclass; ++pj) 
      newprobs[pk][pi][pj] = 0;
  
  //  obain the evidences
  if ((fp1=fopen( fimage1, "r")) == NULL)
     errors("file_hard doesn't exist\n"); 
  if ((fp2=fopen( fimage2, "r")) == NULL)
     errors("file_ve doesn't exist\n"); 

  // Iteration
  double 	lc=0.0;
  double 	lf=0.0;  
  for (i=0; i<num_image; ++i) {
    //printf("image: %d\n", i);
      // input
      for (j=0; j<nsize; ++j) {
          fscanf(fp1, "%f ", &temp);
          images1[j] = temp;
          for (k=0; k<nclass; ++k) 
              fscanf(fp2, "%f ", &images2[j][k]);
      }

      ChangeFormat( images2, labels2 );
      ChangeFormat( images1, labels1 );
      // set initial value for scale_d 
      tree->SetInitialValue();
      // set hard evidence
      tree->SetEvidence( labels1 );
      RunQuadtree( probs_out1, &temp );
      fval[i+1] = -temp;
      lc += -temp;

      // set initial value for scale_d 
      tree->SetInitialValue();
      // set virtual evidence
      tree->SetVirtualEvidence( labels2 );
      RunQuadtree( probs_out2, &temp);
      fval[i+1] += temp;
      lf += -temp;
      // estiamte 
      for (pk=0; pk<ncpt; ++pk)
      for (pi=0; pi<nclass; ++pi)
      for (pj=0; pj<nclass; ++pj)
          newprobs[pk][pi][pj] += 
          (probs_out1[pk][pi][pj]-probs_out2[pk][pi][pj]);
  }

  fclose(fp1);
  fclose(fp2);
  // Calculate the function and gradients
  // probs_out1 -> m
  // probs_out2 -> n
 
  // function value
  fval[0] = (lc -lf)/(double)num_image;
  
  // calculate the gradients: w.r.t z
  for (pk=0; pk<ncpt; ++pk)
  for (pi=0; pi<nclass; ++pi) {
      temp = 0;
      for (pj=0; pj<nclass; ++pj) 
          temp += newprobs[pk][pi][pj] ;
      for (pj=0; pj<nclass; ++pj) {
          grad[pk][pi][pj] = (-newprobs[pk][pi][pj]+
                             probs[pk][pi][pj] * temp)/
                             (float)num_image;
      }
  }
   
  delete newprobs;
  delete probs_out1;
  delete probs_out2;
 } // end of TrainTSBN::CalFunctionGradient()



////////////////////////////////////////////////////////////
//
// void    updateProbs(int      nrow, 
//                     int      ncol, 
//                     int      ndep, 
//                     float    alpha, 
//                     float*** probs_old, 
//                     float*** d, 
//                     float*** probs)
// DESCRIPTION
// Update CTPs by using unconstrained auxiliary variables
//
////////////////////////////////////////////////////////////
void           updateProbs(int      nrow, 
                           int      ncol, 
                           int      ndep, 
                           float    alpha, 
                           float*** probs_old, 
                           float*** d, 
                           float*** probs)
{
 int    i,j,k;
 double den;

 for (i=0; i<nrow; ++i) 
 for (j=0; j<ncol; ++j) {
     den = 0.0;
     for (k=0; k<ndep; ++k) {
         probs[i][j][k] = probs_old[i][j][k]*
                          exp((double)(alpha*d[i][j][k]));
         den += probs[i][j][k];
     }
     if ((den > ZERO) && (den<DBL_MAX))
        for (k=0; k<ndep; ++k) 
            probs[i][j][k]  /= den;
     else
       // errors("updateProbs: denomanator is zero or inf.\n");
       if ( (i>0) || (j==0) )
       printf("updateProbs: denomanator is %f\n", den);
 }
} // end of updateProbs()

         
         
         

//////////////////////////////////////////////////////////
//
// void  TrainTSBN::CheckGradient( char*    fimage1, 
//                             char*    fimage2, 
//                             float*** probs)
// DESCRIPTION
// To check the gradients for the CML training.
//
///////////////////////////////////////////////////////////
void  TrainTSBN::CheckGradient( char*    fimage1, 
                                 char*    fimage2, 
                                 float*** probs)
{
 int		i, j, k, n, m, pi,pj,pk;		
 s_int          nlevel = tree->num_levels;
 s_int          nrow = tree->num_rows[nlevel-1];
 s_int          ncol = tree->num_cols[nlevel-1];
 s_int          ncpt = tree->num_cpts;
 s_int          nclass = tree->num_classes;
 double	 	eps = 1e-4;
 double** 	f = dmatrix(nrow*ncol*ncpt,2);
 float***       grad = fmatrix(ncpt,nclass, nclass);
 float***       probs_val = fmatrix(ncpt, nclass, nclass);
 float***       deltaz = fmatrix(ncpt, nclass, nclass);
 double*        fval = new double[num_image+1];
 float          temp;
 FILE*          fp;

 printf("Checking gradient ....");
 n=0;
 for (k=0; k<ncpt; ++k)
 for (i=0; i<nclass; ++i)
 for (j=0; j<nclass; ++j) {
     printf("n=%d \n",n);
     if ((k==0)&&(i>0)) {
        ++n;
        continue;
     }
     for (pk=0; pk<ncpt; ++pk)
     for (pi=0; pi<nclass; ++pi)
     for (pj=0; pj<nclass; ++pj)
         deltaz[pk][pi][pj] = 0.0;
     deltaz[k][i][j] = eps;
     updateProbs(ncpt, nclass, nclass, (float)1.0,
                 probs, deltaz,  probs_val);
     CalFunctionGradient( fimage1, fimage2, 
                          probs_val, fval,grad );
     f[n][0] = fval[0];
     deltaz[k][i][j] -= 2*eps; 
     updateProbs(ncpt, nclass, nclass, (float)1.0,
                 probs, deltaz,  probs_val);
     CalFunctionGradient( fimage1, fimage2, probs_val, 
                          fval,grad );
     f[n][1] = fval[0]; 
     ++n;
 }    

 CalFunctionGradient( fimage1, fimage2, probs, fval,grad );

 double  deltaf;
 fp = fopen( "CheckGradient.dat", "w");
 n=0;
 for (k=0; k<ncpt; ++k)
 for (i=0; i<nclass; ++i)
 for (j=0; j<nclass; ++j) {
     deltaf = 0.5*(f[n][0]-f[n][1])/eps;
     fprintf(fp, "grad=%e  deltaf=%e: diff=%e\n", 
             grad[k][i][j], deltaf, grad[k][i][j]-deltaf);
     ++n;
 }

 fclose(fp);
 printf("\n");

} // en of TrainTSBN::CheckGradient 
   




//////////////////////////////////////////////////////////
//
//  Function: CML training using Scaled Conjugate gradient.
//  void  TrainTSBN::TrainCMLscgFast( char*    fimage1, 
//			          char*    fimage2, 
//				  float*** probs, 
//                                double*  options)
//  Description 
//  Using a scaled conjugate gradients algorithm to find a 
//  local maximum of the conditional log likelihood 
//  function of the TSBN given hard_evidence 
//  ( stored in file: fimage1) and virtual evidence 
//  (stored in file: fimage2). The initial CPTs are given 
//  in 3D array probs.  The estimated CPTs are returned as 
//  probs. The optional parameters have the following
//  interpretations.
//
//  OPTIONS[0] is set to one to display error values; 
//
//  OPTIONS[1] is a measure of the absolute precision 
//             required for the value of PROBS at the 
//             solution. If the absolute difference 
//             between the values of the PROBS between 
//             two successive steps is less than OPTIONS[1],
//             then the condition is satisfied.
//
//  OPTIONS[2] is a measure of the precision required of 
//             the objective function at the solution. If 
//             the absolute difference between the objective 
//             function values between two successive steps 
//             is less than OPTIONS[2], then this condition 
//             is satisfied. Both this and the previous 
//             condition must be satisfied for termination.
//
//  OPTIONS[3] is the tolerance for stopping the 
//         training.
//
//  OPTIONS[7] returns the conditional log likilihood 
//             values at the optimum values.
//
//  OPTIONS[9] returns the total number of function 
//             evaluations 
//             (including those in any line searchs).
//
//  OPTIONS[10] returns the total number of gradient 
//              evaluations.
//
//  OPTIONS[13] is the maximum number of Iterations.
//
//  OPTIONS[14] is set to 1 to check the gradients
//
//
//  See also
//  train.cc
//
///////////////////////////////////////////////////////////
void   TrainTSBN::TrainCMLscgFast( s_int*   fimage1, 
			           float**  fimage2, 
			           float*** probs, 
                                   float*   options)
{
 int        nclass = tree->num_classes;
 int        ncpt = tree->num_cpts;
 s_int      nrow = tree->num_rows[tree->num_levels-1];
 s_int      ncol = tree->num_cols[tree->num_levels-1];
 int        nsize = nrow*ncol;
 char       fname[SIZE_FILENAME];
 s_int      i, j, k;
 s_int*     he = ChangeFormat( fimage1 );
 float**    ve = ChangeFormat( fimage2 );
 float**    logs = fmatrix(options[13], (num_image+1));
 FILE	    *fp, *fp1, *fp2;

 //  Initial some paramters
 s_int	    success =1;
 int        nsuccess = 1;// nsuccess counts number
                         // of succes.
 int        niter = 1;   // niter counts number of 
                         // Iterations.
 double     sigma0 = 1.0e-4; 
 double     lambda = 1.0;// Initial scale parameter.
 double     lambdamin = 1.0e-15; 
 double     lambdamax = 1.0e100;  
 float***   probs_val = fmatrix( ncpt, nclass, nclass );
 //  Initial function and gradient values
 double*    fold = new double[num_image+1];
 double*    fnew = new double[num_image+1];
 double*    f_temp = new double[num_image+1]; 
 float***   grad = fmatrix(ncpt,nclass, nclass);
 float***   gplus = fmatrix(ncpt,nclass, nclass);
 float***   gold  = fmatrix(ncpt,nclass, nclass);
 float***   gtemp = fmatrix(ncpt,nclass, nclass);
 void       updateProbs(int, int, int, float, 
                        float***, float***, float***);
 
 CalFunctionGradientFast( he, ve, probs, fold, grad );
 ++options[10];
 for (i=0; i<=num_image; ++i)
      logs[0][i] = fold[i];
 // Initial search direction
 float***      srch = fmatrix(ncpt,nclass, nclass);
 setArray3D(ncpt, nclass, nclass, (float)-1.0, grad, srch);
 setArray3D(ncpt, nclass, nclass, (float)1.0, 
            probs, probs_val);
 // Main optimization loop
 double		mu;
 double		sigma;
 double		kappa;
 double         alpha;
 double         gamma;
 double         delta;
 double 	rho;
 double         beta;
 double         temp;
 do{
    // Calculate first and second directional derivatives
    if ( success == 1) {
       mu = productArray3D( ncpt, nclass, nclass, 
                            srch, grad);
       if (mu >= 0.0) {
          setArray3D( ncpt, nclass, nclass, 
                      (float)-1.0, grad, srch);
          mu = productArray3D( ncpt, nclass, nclass, 
                               srch, grad);
       }
       kappa = productArray3D( ncpt, nclass, nclass, 
                               srch, srch);
       sigma = sigma0 / sqrt(kappa);
       updateProbs(ncpt, nclass, nclass, (float)sigma, 
                   probs_val, srch, probs);
       CalFunctionGradientFast( he, ve, probs, 
                                f_temp, gplus );  
       ++options[10];  
       // gamma = (srch * (gplus'-grad'))/sigma
       gamma = (double)productArray3D(ncpt, nclass, 
                                      nclass, srch, gplus);
       gamma -= (double)productArray3D(ncpt, nclass, 
                                       nclass, srch, grad);
       gamma /= sigma;
    } // if success == 1

    // Increase effective curvature and evaluate 
    // step size alpha
    delta = gamma + lambda * kappa;
    if (delta <= 0.0) {
       delta = lambda*kappa;
       lambda -= 2*gamma/kappa;
    }
    alpha = -mu/delta;
    
    // Calculate the comparison ratio
    updateProbs(ncpt, nclass, nclass, (float)alpha,
                probs_val, srch, probs);
    CalFunctionGradientFast( he, ve, probs, 
                         fnew, gtemp );
    ++options[9];
    rho = 2 * (fnew[0]-fold[0]) / (alpha*mu);
    if (rho >= 0.0) success = 1;
    else success = 0;

    // Update the parameters to new location
    if (success == 1) {
       updateProbs(ncpt, nclass, nclass, (float)alpha, 
                   probs_val, srch, probs_val);
       ++nsuccess; 
    }
    setArray3D(ncpt, nclass, nclass, (float)1.0, 
               probs_val, probs);
    
    // display
    if (options[0] == 1) 
       printf("Iter: %4d  Error: %e  Scale: %e\n", 
              niter, fold[0], lambda);
    //
    if (success == 1) {
       // Test for termination 
       f_temp[0] = 0.0;
       for (k=0; k<ncpt; ++k)
       for (i=0; i<nclass; ++i)
       for (j=0; j<nclass; ++j)
           if ( f_temp[0] < fabs(srch[k][i][j]) )
              f_temp[0] = fabs(srch[k][i][j]);    
       if ((alpha*f_temp[0]<options[1]) && 
           (fabs(fnew[0]-fold[0])<options[2])) {
          options[7] = fnew[0];
          printf("Conditions are satistified.\n");
          return;
       }
       else {
	  // Update variables for new position
	  //fold = fnew;
          for (i=0; i<=num_image; ++i)  fold[i] = fnew[i];
          setArray3D(ncpt, nclass, nclass, 
                     (float)1.0, grad, gold);
          setArray3D(ncpt, nclass, nclass, 
                     (float)1.0, gtemp, grad);
          ++options[10];
          // If the gradient is zero then we are done.
          temp = (double)productArray3D(ncpt, nclass, 
                                        nclass, grad, grad);
          if (temp <= ZERO ){
             options[7] = fnew[0];
             printf("The gradient is zero.\n");
             return;
          }
       }
    } // end if success === 1

    // Adjust lambda according to comparison ratio
    if (rho < 0.25) {
       lambda = 4.0 * lambda;
       if (lambda > lambdamax) lambda = lambdamax;
    }
    else if (rho > 0.75) {
       lambda = 0.5*lambda;
       if (lambda < lambdamin) lambda = lambdamin;
    }

    // Re-compute search direction using Hestenes-Steifel 
    // formula, or re-start in direction of negative 
    // gradient after nparams steps.
    if (nsuccess == (ncpt*nclass*nclass)) {
       setArray3D(ncpt, nclass, nclass, 
                  (float)-1.0, grad, srch);
       nsuccess = 0;
    }
    else if (success == 1) { 
            beta  = productArray3D(ncpt, nclass, 
                                   nclass, gold, grad);
            beta -= productArray3D(ncpt, nclass, 
                                   nclass, grad, grad);
            beta /= mu;
            sumArray3D(ncpt, nclass, nclass, 
                       (float)-1.0, (float)beta, 
                       grad, srch, srch);
    }
    
    // Output the probs and fval
    for (i=0; i<=num_image; ++i)
        logs[niter][i] = fold[i];
    
    fp = fopen("logsFast.dat", "w");
    for (i=0; i<=niter; ++i) {
        for (j=0; j<=num_image; ++j)
            fprintf(fp, "%f ", -logs[i][j]);
        fprintf(fp, "\n");
    }
    fclose(fp);
    
    sprintf(fname, "probsFast%d.dat", niter);
    fp = fopen(fname, "w");
    for (j=0; j<nclass; ++j)
        fprintf(fp, "%f ", probs[0][0][j]);
    fprintf(fp, "\n");
    for (k=1; k<ncpt; ++k) 
    for (i=0; i<nclass; ++i) {
        for (j=0; j<nclass; ++j) 
            fprintf(fp, "%f ", probs[k][i][j]);
        fprintf(fp, "\n");   
    }       
    fclose(fp);


    ++niter;            
 } while (niter < options[13]);
  
 // If we get here, then we havn't terminated in the 
 // given number of Iterations.
 options[7] = fold[0];
 if (options[0] >= 0)
    printf("Warning: Maximum number of Iterations has\
            been exceeded.\n");
 
} // end of TrainTSBN::TrainCMLscg()



//////////////////////////////////////////////////////////
//
// void TrainTSBN::CalFunctionGradientFast(char*    fimage1,
//				       char*    fimage2, 
//				       float*** probs, 
//				       double*  fval,
//				       float*** grad )	
// Description:
//
//////////////////////////////////////////////////////////
void TrainTSBN::CalFunctionGradientFast(s_int*   he,
				        float**  ve, 
				        float*** probs, 
				        double*  fval,
                                        float*** grad )
{
  int 	    i, j,k,pi,pj,pk;
  s_int     nlevel = tree->num_levels;
  s_int     nrow = tree->num_rows[nlevel-1];
  s_int     ncol = tree->num_cols[nlevel-1];
  s_int     ncpt = tree->num_cpts;
  s_int     nclass = tree->num_classes;
  int	    nsize = nrow*ncol;
  float     temp;

  // set the priors and CPT to b.net
  //tree->SetCPT1( probs ); // One CPT on each level
  tree->SetCPT6( probs ); // six CPT on level 1
  // 
  float***      probs_out1 = fmatrix(ncpt, nclass, nclass);
                                     // m^n
  float***      probs_out2 = fmatrix(ncpt, nclass, nclass);
                                     // n^n
  double***     newprobs = dmatrix(ncpt, nclass, nclass);  
                                     // sum(m^n-n^n)
  for (pk=0; pk<ncpt; ++pk)
  for (pi=0; pi<nclass; ++pi)
  for (pj=0; pj<nclass; ++pj) 
      newprobs[pk][pi][pj] = 0;

  // Iteration
  double 	lc=0.0;
  double 	lf=0.0;  
  for (i=0; i<num_image; ++i) {
    //printf("image: %d\n", i);
 
      // set initial value for scale_d 
      tree->SetInitialValue();
      // set hard evidence
      tree->SetEvidence( (he+i*nrow*ncol) );
      RunQuadtree( probs_out1, &temp );
      fval[i+1] = -temp;
      lc += -temp;

      // set initial value for scale_d 
      tree->SetInitialValue();
      // set virtual evidence
      tree->SetVirtualEvidence( (ve+i*nrow*ncol) );
      RunQuadtree( probs_out2, &temp);
      fval[i+1] += temp;
      lf += -temp;
      // estiamte 
      for (pk=0; pk<ncpt; ++pk)
      for (pi=0; pi<nclass; ++pi)
      for (pj=0; pj<nclass; ++pj)
          newprobs[pk][pi][pj] += 
          (probs_out1[pk][pi][pj]-probs_out2[pk][pi][pj]);
  }
  // Calculate the function and gradients
  // probs_out1 -> m
  // probs_out2 -> n
 
  // function value
  fval[0] = (lc -lf)/(double)num_image;
  
  // calculate the gradients: w.r.t z
  for (pk=0; pk<ncpt; ++pk)
  for (pi=0; pi<nclass; ++pi) {
      temp = 0;
      for (pj=0; pj<nclass; ++pj) 
          temp += newprobs[pk][pi][pj] ;
      for (pj=0; pj<nclass; ++pj) {
          grad[pk][pi][pj] = (-newprobs[pk][pi][pj]+
                             probs[pk][pi][pj] * temp)/
                             (float)num_image;
      }
  }
   
  delete newprobs;
  delete probs_out1;
  delete probs_out2;
 } // end of TrainTSBN::CalFunctionGradient()





