#include <math.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <iostream.h>
#include "util.hh"

// allocate a short integer matrix with nrow rows and ncol 
// columns
s_int **simatrix(int nrow, int ncol)
{
  s_int	**m, i;

  m = (s_int **)malloc((size_t)nrow*sizeof(s_int*));
  if (!m) 
     fprintf(stderr, "allocation failure 1 in simatrix()");
  m[0]= (s_int *) malloc((size_t)nrow*ncol*
        sizeof(s_int));
  for( i=1; i<nrow; i++ )  m[i]= m[i-1] + ncol;
  return m;
}


// allocate a unsigned char matrix with nrow rows and ncol 
// columns
char  **cmatrix(int nrow, int ncol)
{
  char	**m;
  int   i;

  m = (char **)malloc((size_t)nrow*sizeof(char*));
  if (!m) 
     fprintf(stderr, "allocation failure 1 in ucmatrix()");
  m[0]= (char *)malloc((size_t)nrow*ncol*sizeof(char));
  for( i=1; i<nrow; i++ )  m[i]= m[i-1] + ncol;
  return m;
}


// allocate a integer matrix with nrow rows and ncol columns
int **imatrix(int nrow, int ncol)
{
  int	**m, i,j;

  m = (int **)malloc((size_t)nrow*sizeof(int*));
  if (!m) 
     fprintf(stderr, "allocation failure 1 in imatrix()");
  m[0]= (int *)malloc((size_t)nrow*ncol*sizeof(int));
  for( i=1; i<nrow; i++ )  m[i]= m[i-1] + ncol;


  for (i=0; i<nrow; ++i)
  for (j=0; j<ncol; ++j)  m[i][j]=0;

  return m;
}


// allocate a integer vector with n elements
int *ivector(int n)
{
  int	*m;

  m = (int *)malloc((size_t)n*sizeof(int));
  if (!m) 
     fprintf(stderr, "allocation failure 1 in ivector()");
  return m;
}



////////////////////////////////////////////////////////////
//
// Aim: allocate a float  matrix
//
/////////////////////////////////////////////////////////// 
float **fmatrix(int nrow, int ncol)
{
 int      i, j;
 float** m;

 m = new float * [nrow];
 for(i=0; i< nrow;i++)   m[i] = new float [ncol];

 for (i=0; i<nrow; ++i)
 for (j=0; j<ncol; ++j) m[i][j]=0;

 return m;
}



float  ***fmatrix(int nrow, int ncol, int ndep)
{
 int	i, j;
 float ***m;

 m = new float ** [nrow];
 for (i=0; i<nrow; ++i) {
     m[i] = new float * [ncol];
     for (j=0; j<ncol; ++j) 
         m[i][j] = new float [ndep];
 }

 return m;
} 




///////////////////////////////////////////////////////////
//
// double **dmatrix(int nrow, int ncol)
// Aim: allocate a double matrix with nrow rows and 
//      ncol columns
//
/////////////////////////////////////////////////////////// 
double **dmatrix0(int nrow, int ncol)
{
 int      i, j;
 double** m;

 m = new double * [nrow];
 for(i=0; i< nrow;i++)   m[i] = new double [ncol];

 for (i=0; i<nrow; ++i)
 for (j=0; j<ncol; ++j) m[i][j]=0;

 return m;
}


double **dmatrix(int nrow, int ncol)
{
  int	 i, j;
  double **m;

  m = (double **)malloc((size_t)nrow*ncol*sizeof(double*));
  if (!m) 
     fprintf(stderr, "allocation failure 1 in dmatrix()");
  for (i=0; i<nrow; ++i)
      m[i] = (double *)malloc((size_t)ncol*sizeof(double));

  for (i=0; i<nrow; ++i)
  for (j=0; j<ncol; ++j) m[i][j]=0;

  return m;
}


double ***dmatrix(int nrow, int ncol, int ndep)
{
 int	i, j, k;
 double ***m;

 m = (double ***) malloc( (size_t)(nrow*ncol*ndep
                          *sizeof(double**)) );
 if (!m) {
   cout << "allocation failure 1 in dmatrix()!\n";
   exit(1);
 }
 for (i=0; i<nrow; ++i) {
     m[i] = (double**) malloc( (size_t)(ncol*ndep
                               *sizeof(double*)) );
     for (j=0; j<ncol; ++j)
         m[i][j] = (double*) malloc( (size_t)(ndep
                               *sizeof(double)) );
 }

 for (i=0; i<nrow; ++i)
 for (j=0; j<ncol; ++j)
 for (k=0; k<ndep; ++k) m[i][j][k] = 0;

 return m;

} // end of dmatrix;


// allocate a double vector with n elements
double *dvector(int n)
{
  double *m;

  m = (double *)malloc( (size_t)n*sizeof(double) );
  if (!m) 
     fprintf(stderr, "allocation failure 1 in ivector()");
  return m;
}


int  imin(int a, int b)
{ 
 if (a >= b) return b;
 else return a;
}


///////////////////////////////////////////////////////////
// void  setArray3D(nrow, ncol, ndep, alpha, 
//                  arrayLeft, arrayRight)
//
// DESCRIPTION
// arrayL = alpha * arrayR
//
// 25/05/98
///////////////////////////////////////////////////////////
void	setArray3D(int      nrow, 
                   int      ncol, 
                   int      ndep, 
                   float    alpha,
                   float*** arrayR, 
                   float*** arrayL)
{
 int	i, j, k;
 for (i=0; i<nrow; ++i)
 for (j=0; j<ncol; ++j)
 for (k=0; k<ndep; ++k)
     arrayL[i][j][k]= alpha*arrayR[i][j][k];
}


///////////////////////////////////////////////////////////
// void  sumArray3D(nrow, ncol, ndep, alpha, beta,
//                  arrayLeft1, arrayLeft2,  arrayRight)
//
// DESCRIPTION
// arrayL = alpha * arrayR1 + beta * arrayR2
//
// 25/05/98
///////////////////////////////////////////////////////////
void 	sumArray3D(int nrow, 
                   int ncol, 
                   int ndep, 
                   float alpha, 
                   float beta,
                   float*** arrayR1,  
                   float*** arrayR2,
                   float*** arrayL)
{
 int	i, j, k;

 for (i=0; i<nrow; ++i)
 for (j=0; j<ncol; ++j)
 for (k=0; k<ndep; ++k)
     arrayL[i][j][k]= alpha*arrayR1[i][j][k]+
                      beta*arrayR2[i][j][k];
}



//////////////////////////////////////////////////////////
// double  setArray3D(nrow, ncol, ndep, alpha, beta,
//                  arrayLeft1, arrayLeft2,  arrayRight)
//
// DESCRIPTION
// prod =  arrayLeft1 * arrayLeft2'
//
// 25/05/98
///////////////////////////////////////////////////////////
double 	productArray3D(int      nrow, 
                       int      ncol, 
                       int      ndep, 
                       float*** arrayR1,  
                       float*** arrayR2)
{
 int	i, j, k;
 double  prod=0.0;

 for (i=0; i<nrow; ++i)
 for (j=0; j<ncol; ++j)
 for (k=0; k<ndep; ++k)
     prod += arrayR1[i][j][k] * arrayR2[i][j][k];
 return  prod;
}



///////////////////////////////////////////////////////////
//
//  double ran1(long*)
//
//  C M Bishop
//  3 November 1993
//
//  This function generates uniform random numbers in 
//  the range (0,1).
//  It is taken from Numerical Recipes in C, Version 2.0.
//
///////////////////////////////////////////////////////////


#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-13
#define RNMX (1.0-EPS)

double ran1(long *idum)
{
	int j;
	long k;
	static long iy=0;
	static long iv[NTAB];
	double temp;

	if (*idum <= 0 || !iy) {
		if (-(*idum) < 1) *idum=1;
		else *idum = -(*idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(*idum)/IQ;
			*idum=IA*(*idum-k*IQ)-IR*k;
			if (*idum < 0) *idum += IM;
			if (j < NTAB) iv[j] = *idum;
		}
		iy=iv[0];
	}
	k=(*idum)/IQ;
	*idum=IA*(*idum-k*IQ)-IR*k;
	if (*idum < 0) *idum += IM;
	j= int(iy/NDIV);
	iy=iv[j];
	iv[j] = *idum;
	if ((temp=AM*iy) > RNMX) return RNMX;
	else return temp;
}

//////////////////////////////////////////////////////////
//
//  double gasdev(long*)
//
//  This function generates a random number from a gaussian 
//  distribution with zero mean and unit standard deviation.
//  It is taken from Numerical Recipes in C, Version 2.0.
//
///////////////////////////////////////////////////////////
double gasdev(long *idum)
{
	double ran1(long *idum);
	static int iset=0;
	static double gset;
	double fac,rsq,v1,v2;

	if  (iset == 0) {
		do {
			v1=2.0*ran1(idum)-1.0;
			v2=2.0*ran1(idum)-1.0;
			rsq=v1*v1+v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac=sqrt(-2.0*log(rsq)/rsq);
		gset=v1*fac;
		iset=1;
		return v2*fac;
	} else {
		iset=0;
		return gset;
	}
}


///////////////////////////////////////////////////////////
//  double RandNoGen:GetGauss(double, double)
//
//
//  This function is used to generate a double from 
//  a gaussian distribution with mean m and standard 
//  deviation s.
//
///////////////////////////////////////////////////////////
double Getgauss(double m, double s, long* seed)
{
  double gasdev(long *idum);
  return m + s*gasdev(&(*seed));
}


//////////////////////////////////////////////////////////
// Access a file
//////////////////////////////////////////////////////////
FILE* file_open(char* fname, char* mode)
{
 FILE*	fp;
 
 fp = fopen(fname, mode);
 if (!fp)  cout << "Error in opening  file: "<<fname;
 return fp;
}




////////////////////////////////////////////////////////////
// print error messages
////////////////////////////////////////////////////////////
void  errors(char* msg)
{
  cout << msg <<endl;
  flush(cout);
  exit(1);
}
