/////////////////////////////////////////////////////////////
//
// Filename:	train.cc
// Description: 
//
//
// Copyright (c) 1998 Xiaojuan Feng 
//
/////////////////////////////////////////////////////////////
#include <math.h>
#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <fstream.h>
#include <stdio.h>

#include "trainTSBN.hh"

int  main(int argc, char** argv)
{
 int		i,j,k;
 FILE		*fp, *fp1;

 // Examples:
 // CML training for real images:
 // train b 61 init_prob.dat 
 //                imagesTr.dat, imagesMlpTrVE.dat
 // CML fast training for real images:
 // train f 61 init_prob.dat 
 //                imagesTr.dat, imagesMlpTrVE.dat
 // ML training for real images:
 // train h 61 init_prob.dat 
 //                imagesTr.dat, 
 if ( (argc != 5)&&(argc != 6) )
    errors("Usage: \
            type \
            num_image \
            file_CPTs \
            havd_evidences \
            virtual_evidences \n");

 // obtain parameters from inputs
 s_int nclass = NUMBER_CLASSES;
 s_int nlevel = NUMBER_LEVELS;  
 s_int nrow   = NUMBER_ROWS;    
 s_int ncol   = NUMBER_COLS;    
 s_int ncpt   = NUMBER_CPTS;    
 s_int nimage = atoi( argv[2] );
 int   nsize  = nrow*ncol*nimage;

 // obtain the initial CPTs
 // format: 
 // first row: priors
 // the first CPT between root node to level 1
 // the second CPT between Level 1 to Level 2
 // ...
 if ((fp=fopen( argv[3], "r")) == NULL)
    errors("file_CPTs doesn't exist\n"); 
 float*** probs = fmatrix(ncpt, nclass, nclass);
 for (i=0; i<nclass; ++i)
     fscanf(fp, "%f", &probs[0][0][i]);
 for (k=1; k<ncpt; ++k)
 for (i=0; i<nclass; ++i)
 for (j=0; j<nclass; ++j)
      fscanf(fp, "%f", &probs[k][i][j]);
 fclose(fp);


 // Set information about the structure of the B.net
 TrainTSBN*  net = new TrainTSBN( nclass, nlevel, nrow, 
                                  ncol,ncpt, nimage );

 // Training by using hard and virtual evidences and 
 // different trainig methods
 float		tol = 1.0e-20;
 float		eta = 0.00005;
 float		temp;
 float		options[15];
 float**		imagesv;
 s_int		maxiter =500;
 s_int*		images;

 switch (argv[1][0]){
 case 'h':
 	// obtain the images
 	if ((fp=fopen( argv[4], "r")) == NULL)
	    errors("file_images doesn't exist\n");
        images = new s_int [nsize];
 	for (i=0; i<nsize; ++i) {
 	    fscanf(fp, "%f", &temp);
 	    images[i] = (s_int)temp;
        }
        fclose(fp);
 	// Hard evidence training 
 	net->TrainML( images, probs, tol, maxiter);
	break;
 case 'v':
 	// obtain the images
 	if ((fp=fopen( argv[4], "r")) == NULL)
	    errors("file_images doesn't exist\n");
        imagesv = fmatrix(nsize, nclass);
 	for (i=0; i<nsize; ++i) 
	for (j=0; j<nclass; ++j) 
 	    fscanf(fp, "%f", &imagesv[i][j]);
        fclose(fp);
        // Virtual evidence training
        net->TrainML(imagesv, probs, tol, maxiter);
	break;

 case 'b':

        // Hard and virtual evidences trainning CML
        options[0] = 1;  // display the error messages.
        options[1] = 1.0e-20;
        options[2] = 1.0e-20; 
        options[3] = tol;   
        options[13] = 500;
        options[14] = 0; // check gradient
        net->TrainCMLscg( argv[4], argv[5], 
                          probs, options );      
	break;

 case 'f':
        // Obtain the images
        if ((fp=fopen( argv[4], "r")) == NULL)
	    errors("file_hard_images doesn't exist\n");
        if ((fp1=fopen( argv[5], "r")) == NULL)
	    errors("file_hard_images doesn't exist\n");
        images = new s_int [nsize];
        imagesv = fmatrix(nsize, nclass);
 	for (i=0; i<nsize; ++i) {
	    fscanf(fp, "%f", &temp);
 	    images[i] = (s_int)temp;            
    	    for (j=0; j<nclass; ++j) 
 	        fscanf(fp1, "%f", &imagesv[i][j]); 
        } 
        fclose(fp);
        fclose(fp1);        
        // Hard and virtual evidences trainning CML
        options[0] = 1;  // display the error messages.
        options[1] = 1.0e-20;
        options[2] = 1.0e-20; 
        options[13] = 2000;
        options[14] = 0; // check gradient
        net->TrainCMLscgFast( images, imagesv, 
                          probs, options );      
	break;
  
 default:
	printf("The first argument should be:\n \
                'h'->hard evidence;\n \
		'v'->virtual evidence; \n\
		'b'->both evidences;\n \
                'f'->both evidences, fast training\n");
        break;
 } // end of switch

 


} // end of main 
